'''Construct and verify an SCT for a certificate from the add-chain response of
a Certificate Transparency (CT) log (cf.
https://tools.ietf.org/html/rfc6962#section-4.1).
'''

# -*- coding: utf-8 -*-

import argparse
import json
import logging

from utlz import flo, first_paragraph, load_json, namedtuple

from ctutlz.utils.logger import loglevel, setup_logging
from ctutlz.ctlog import Log
from ctutlz.sct.sct import _Sct
from ctutlz.sct.ee_cert import EndEntityCert, IssuerCert
from ctutlz.sct.signature_input import create_signature_input_precert
from ctutlz.sct.verification import verify_signature, SctValidationResult
from ctutlz.utils.encoding import decode_from_b64


# TODO FIXME: refactor code
# there should be a signature_input(.. ct_log_resp ..) function
def sct_from_log_response(ct_log_response):
    data_dict = {
        'der': None,

        # 'log_id': decode_from_b64(ct_log_response.id.encode('ascii')),
        'log_id': decode_from_b64(ct_log_response.id),
        'extensions': None,
        'signature_alg_hash': int(0),  # doesn't matter
        'signature_alg_sign': int(0),  # doesn't matter
        'signature_len': None,
        # 'signature': bytes(ct_log_response.signature, 'ascii'),
        'signature': decode_from_b64(ct_log_response.signature),

        # required for signature input creation
        'version': int(ct_log_response.sct_version),
        'timestamp': int(ct_log_response.timestamp),
        'extensions_len': len(ct_log_response.extensions),
    }
    return _Sct(**data_dict)


CtLogResponse = namedtuple(
    typename='CtLogResponse',
    field_names=[
        'extensions',
        'id',
        'sct_version',
        'signature',
        'timestamp',
    ]
)


def create_and_validate_sct_of_precert(ee_cert, issuer_cert, ct_log_response,
                                       log_key_str_b64):
    sct = sct_from_log_response(ct_log_response)

    sign_input_func = create_signature_input_precert

    log = Log(
        description=None,
        key=log_key_str_b64,
        url=None,
        maximum_merge_delay=None,
        operated_by=[''],
    )

    # TODO DEBUG
    print(log.pubkey)
    print(log.pubkey.encode('ascii'))
    signature_input = sign_input_func(ee_cert, sct, issuer_cert)
    open('charon/signature_input.bin', 'wb').write(signature_input)
    import base64

    verified, output, cmd_res = verify_signature(
        signature_input=sign_input_func(ee_cert, sct, issuer_cert),
        signature=sct.signature,
        pubkey=log.pubkey.encode('ascii')
    )
    return SctValidationResult(ee_cert, sct, log, verified, output, cmd_res)


def create_and_validate_sct_of_cert(ee_cert, ct_log_response, log_key_str_b64):
    raise(Exception('not implemented'))


def create_parser():
    parser = argparse.ArgumentParser(description=first_paragraph(__doc__))

    meg = parser.add_mutually_exclusive_group()
    meg.add_argument('--short',
                     dest='loglevel',
                     action='store_const',
                     const=logging.WARNING,
                     default=logging.INFO,  # default loglevel if nothing set
                     help='show short result and warnings/errors only')
    meg.add_argument('--debug',
                     dest='loglevel',
                     action='store_const',
                     const=logging.DEBUG,
                     help='show more for diagnostic purposes')

    req = parser.add_argument_group('required arguments')
    req.add_argument('--cert',
                     metavar='<file>',
                     required=True,
                     help='end entity certificate of the chain submitted to '
                          'the CT log encoded as DER or PEM, or a chain '
                          'containing the precert followed by its issuer cert')
    req.add_argument('--resp',
                     nargs='+',
                     metavar='<file>',
                     required=True,
                     help='one or more ct-log add-chain responses '
                          'encoded as JSON objects')
    req.add_argument('--log-key',
                     nargs='+',
                     metavar='<file>',
                     required=True,
                     help='one or more log pubkeys base64 encoded')
    return parser


def main():
    parser = create_parser()
    args = parser.parse_args()
    logger = setup_logging(args.loglevel)
    logger.debug(args)

    with open(args.cert, 'rb') as fh:
        cert_data = fh.read()

    ee_cert_der = cert_data
    issuer_cert_der = None
    try:
        # cert comments may not are ascii-encoded
        chain_pem = cert_data.decode('utf-8')

        # FIXME: refactor
        _, ee_cert_and_rest = chain_pem.split(
            '-----BEGIN CERTIFICATE-----\n',
            1
        )
        ee_cert_b64, rest = ee_cert_and_rest.split(
            '\n-----END CERTIFICATE-----',
            1
        )
        _, issuer_and_rest = rest.split(
            '-----BEGIN CERTIFICATE-----\n',
            1
        )
        issuer_cert_b64, _ = issuer_and_rest.split(
            '\n-----END CERTIFICATE-----',
            1
        )
        ee_cert_der = decode_from_b64(ee_cert_b64)
        issuer_cert_der = decode_from_b64(issuer_cert_b64)
    except ValueError:
        pass  # swallow error, assuming DER

    ee_cert = EndEntityCert(ee_cert_der)

    with open(args.log_key[0], 'r') as fh:
        # type: str
        log_key_str_b64 = fh.read().replace(
            '-----BEGIN PUBLIC KEY-----\n', '').replace(
            '\n-----END PUBLIC KEY-----', '').replace('\n', '').replace(' ', '')

    for filename_response in args.resp:
        response_dict = load_json(filename_response)
        # with open(filename_response, 'rb') as fh:
        #     response_dict = json.load(fh)
        ct_log_response = CtLogResponse(**response_dict)

        if issuer_cert_der:
            print('** precert **')  # TODO DEBUG
            res = create_and_validate_sct_of_precert(
                ee_cert,
                IssuerCert(issuer_cert_der),
                ct_log_response,
                log_key_str_b64
            )
        else:
            print('** cert (no precert) **')  # TODO DEBUG
            res = create_and_validate_sct_of_cert(
                ee_cert,
                ct_log_response,
                log_key_str_b64
            )
        from ctutlz.scripts.verify_scts import show_validation
        show_validation(res)


if __name__ == '__main__':
    main()



'''namedtuple defs which represent the data structures defined in RFC 6962 -
Certificate Transparency.
'''

import struct

from pyasn1.codec.der.encoder import encode as der_encoder
from pyasn1.codec.der.decoder import decode as der_decoder
from pyasn1_modules import rfc5280
from utlz import flo, namedtuple as namedtuple_utlz

from ctutlz.utils.tdf_bytes import TdfBytesParser, namedtuple
from ctutlz.utils.encoding import decode_from_b64, encode_to_b64
from ctutlz.utils.string import to_hex
from ctutlz.sct.ee_cert import tbscert_without_ct_extensions


# tdf := "TLS Data Format" (cf. https://tools.ietf.org/html/rfc5246#section-4)


# 3.1. Log Entries
# https://tools.ietf.org/html/rfc6962#section-3.1


def _parse_log_entry_type(tdf):
    with TdfBytesParser(tdf) as parser:
        parser.read('val', '!H')  # (65535) -> 2 bytes
        return parser.result()


LogEntryType = namedtuple(
    typename='LogEntryType',
    field_names='arg',
    lazy_vals={
        '_parse_func': lambda _: _parse_log_entry_type,

        'val': lambda self: self._parse['val'],

        'is_x509_entry': lambda self: self.val == 0,
        'is_precert_entry': lambda self: self.val == 1,

        '__str__': lambda self: lambda:
            'x509_entry' if self.is_x509_entry else
            'precert_entry' if self.is_precert_entry else
            flo('<unknown log entry type {self.tdf}>'),
    }
)


def _parse_log_entry(tdf):
    with TdfBytesParser(tdf) as parser:
        entry_type = LogEntryType(
            parser.delegate('entry_type', _parse_log_entry_type))

        # parse entry
        if entry_type.is_x509_entry:
            parser.delegate('entry', _parse_x509_chain_entry)
            parser.res['x509_entry'] = parser.res['entry']
        elif entry_type.is_precert_entry:
            parser.delegate('entry', _parse_precert_chain_entry)
            parser.res['precert_entry'] = parser.res['entry']
        else:
            raise Exception(flo('Unknown entry_type: {entry_type}'))

        return parser.result()


LogEntry = namedtuple(
    typename='LogEntry',
    field_names='arg',
    lazy_vals={
        '_parse_func': lambda _: _parse_log_entry,

        'entry_type': lambda self: LogEntryType(self._parse['entry_type']),
        'entry': lambda self:
            ASN1Cert(self._parse['entry'])
                if self.entry_type.is_x509_entry else
            PreCert(self._parse['entry'])
                if self.entry_type.is_precert_entry else
            None,
    }
)


def _parse_asn1_cert(tdf):
    with TdfBytesParser(tdf) as parser:
        parser.read('len1', '!B')
        parser.read('len2', '!B')
        parser.read('len3', '!B')

        der_len = struct.unpack('=I', struct.pack('!4B',
                                                  0,
                                                  parser.res['len1'],
                                                  parser.res['len2'],
                                                  parser.res['len3']))[0]
        parser.res['der_len'] = der_len
        parser.read('der', flo('!{der_len}s'))

        return parser.result()


ASN1Cert = namedtuple(
    typename='ASN1Cert',
    field_names='arg',
    lazy_vals={
        '_parse_func': lambda _: _parse_asn1_cert,

        'der': lambda self: self._parse['der'],
        'pyasn1': lambda self: der_decoder(self.der, rfc5280.Certificate()),
    }
)


def _parse_asn1_cert_list(tdf):
    with TdfBytesParser(tdf) as parser:
        parser.read('len1', '!B')
        parser.read('len2', '!B')
        parser.read('len3', '!B')

        der_list_len = struct.unpack('=I', struct.pack('!4B',
                                                       0,
                                                       parser.res['len1'],
                                                       parser.res['len2'],
                                                       parser.res['len3']))[0]
        der_end_offset = parser.offset + der_list_len

        list_of_parse_asn1_cert = []
        while parser.offset < der_end_offset:
            parse_asn1_cert = parser.delegate(_parse_asn1_cert)
            list_of_parse_asn1_cert.append(parse_asn1_cert)

        parser.res['der_list_len'] = der_list_len
        parser.res['list_of_parse_asn1_cert'] = list_of_parse_asn1_cert

        return parser.result()


ASN1CertList = namedtuple(
    typename='ASN1CertList',
    field_names='arg',
    lazy_vals={
        '_parse_func': lambda _: _parse_asn1_cert_list,

        'certs': lambda self: [
            ASN1Cert(parse_asn1_cert)
            for parse_asn1_cert
            in self._parse['list_of_parse_asn1_cert']
        ],
    }
)


def _parse_x509_chain_entry(tdf):
    with TdfBytesParser(tdf) as parser:
        parser.delegate('leaf_certificate', _parse_asn1_cert),
        parser.delegate('certificate_chain', _parse_asn1_cert_list),
        return parser.result()


X509ChainEntry = namedtuple(
    typename='X509ChainEntry',
    field_names='arg',
    lazy_vals={
        '_parse_func': lambda _: _parse_x509_chain_entry,

        'leaf_certificate': lambda self:
            ASN1Cert(self._parse['leaf_certificate']),
        'certificate_chain': lambda self:
            ASN1CertList(self._parse['certificate_chain']),
    }
)


def _parse_precert_chain_entry(tdf):
    with TdfBytesParser(tdf) as parser:
        parser.delegate('pre_certificate', _parse_asn1_cert),
        parser.delegate('precert_chain', _parse_asn1_cert_list),
        return parser.result()


PrecertChainEntry = namedtuple(
    typename='PrecertChainEntry',
    field_names='arg',
    lazy_vals={
        '_parse_func': lambda _: _parse_precert_chain_entry,

        'pre_certificate': lambda self:
            ASN1Cert(self._parse['pre_certificate']),
        'precertificate_chain': lambda self:
            ASN1CertList(self._parse['precert_chain']),
    }
)


# 3.2 Structure of the Signed Certificate Timestamp
# https://tools.ietf.org/html/rfc6962#section-3.2


def _parse_signature_type(tdf):
    with TdfBytesParser(tdf) as parser:
        parser.read('val', '!B')
        return parser.result()


SignatureType = namedtuple(
    typename='SignatureType',
    field_names='arg',
    lazy_vals={
        '_parse_func': lambda _: _parse_signature_type,

        'val': lambda self: self._parse['val'],

        'is_certificate_timestamp': lambda self: self.val == 0,
        'is_tree_hash': lambda self: self.val == 1,

        '__str__': lambda self: lambda:
            'certificate_timestamp' if self.is_certificate_timestamp else
            'tree_hash' if self.is_tree_hash else
            '<unknown signature type>',
    }
)


def _parse_version(tdf):
    with TdfBytesParser(tdf) as parser:
        parser.read('val', '!B')
        return parser.result()


Version = namedtuple(
    typename='Version',
    field_names='arg',
    lazy_vals={
        '_parse_func': lambda _: _parse_version,

        'val': lambda self: int(self._parse['val']),

        'is_v1': lambda self: self.val == 0,

        '__str__': lambda self: lambda:
            'v1' if self.is_v1 else
            '<unkown version>',
    }
)


def _parse_log_id(tdf):
    with TdfBytesParser(tdf) as parser:
        parser.read('val', '!32s')
        return parser.result()


LogID = namedtuple(
    typename='LogID',
    lazy_vals={
        '_parse_func': lambda _: _parse_log_id,

        # type: int, '!L', [32]
        # https://docs.python.org/3/library/struct.html#format-characters
        'val': lambda self: bytes(self._parse['val']),
    },
)


def _parse_tbs_certificate(tdf):
    with TdfBytesParser(tdf) as parser:
        parser.read('len1', '!B')
        parser.read('len2', '!B')
        parser.read('len3', '!B')
        len_der = struct.unpack('=I', struct.pack('!4B',
                                                  0,
                                                  parser.res['len1'],
                                                  parser.res['len2'],
                                                  parser.res['len3']))[0]
        from_ = parser.offset
        parser.offset += len_der
        until = parser.offset
        parser.res['der'] = tdf[from_:until]
        return parser.result()


TBSCertificate = namedtuple(
    typename='TBSCertificate',
    field_names='arg',
    lazy_vals={
        '_parse_func': lambda _: _parse_tbs_certificate,

        'der': lambda self: bytes(self._parse['der']),
        'pyasn1': lambda self: der_decoder(self.der,
                                           asn1Spec=rfc5280.TBSCertificate()),

        'len': lambda self: len(self.der),
        'lens': lambda self: struct.unpack('!4B', struct.pack('!I', self.len)),
        'len1': lambda self: self.lens[1],
        'len2': lambda self: self.lens[2],
        'len3': lambda self: self.lens[3],

        'without_ct_extensions': lambda self:
            der_encoder(
                TBSCertificate(tbscert_without_ct_extensions(self.pyasn1))),
    }
)


def _parse_pre_cert(tdf):
    with TdfBytesParser(tdf) as parser:
        parser.read('issuer_key_hash', '!32s')
        parser.delegate('tbs_certificate', _parse_tbs_certificate)
        return parser.result()


PreCert = namedtuple(
    typename='PreCert',
    field_names='arg',
    lazy_vals={
        '_parse_func': lambda _: _parse_pre_cert,

        'issuer_key_hash': lambda self: bytes(self._parse['issuer_key_hash']),
        'tbs_certificate': lambda self:
            TBSCertificate(self._parse['tbs_certificate']),
    }
)


def _parse_ct_extensions(tdf):
    with TdfBytesParser(tdf) as parser:
        parser.read('len', '!H')
        parser.res['val'] = None  # "Currently, no extensions are specified"
        return parser.result()


CtExtensions = namedtuple(
    typename='CtExtensions',
    field_names='arg',
    lazy_vals={
        '_parse_func': lambda _: _parse_ct_extensions,

        'len': lambda self: self._parse['len'],
        'val': lambda self: self._parse['val'],
    }
)


def _parse_signed_certificate_timestamp(tdf):
    with TdfBytesParser(tdf) as parser:
        parser.delegate('version', _parse_version)
        parser.delegate('id', _parse_log_id)
        parser.read('timestamp', '!Q')

        parser.delegate('ct_extensions', _parse_ct_extensions)

        # digitally-signed struct
        parser.read('signature_alg_hash', '!B'),
        parser.read('signature_alg_sign', '!B'),
        signature_len = parser.read('signature_len', '!H')
        parser.read('signature', flo('!{signature_len}s'))

        return parser.result()


SignedCertificateTimestamp = namedtuple(
    typename='SignedCertificateTimestamp',
    field_names='arg',
    lazy_vals={
        '_parse_func': lambda _: _parse_signed_certificate_timestamp,

        'version': lambda self: Version(self._parse['version']),
        'id': lambda self: LogID(self._parse['id']),
        'timestamp': lambda self: int(self._parse['timestamp']),
        'extensions': lambda self: CtExtensions(self._parse['ct_extensions']),

        # digitally-signed struct
        # https://tools.ietf.org/html/rfc5246#section-4.7
        'signature_algorithm_hash': lambda self:
            int(self._parse['signature_alg_hash']),
        'signature_algorithm_signature': lambda self:
            int(self._parse['signature_alg_sign']),
        'signature_len': lambda self: int(self._parse['signature_len']),
        'signature': lambda self: bytes(self._parse['signature']),

        'log_id': lambda self: self.id,
        'log_id_b64': lambda self: encode_to_b64(self.log_id.tdf),  # type: str
        'version_hex': lambda self: to_hex(self.version.tdf),
        'timestamp_hex': lambda self: to_hex(self.timestamp),
        'extensions_len': lambda self: self.extensions.len,
        'extensions_len_hex': lambda self: to_hex(self.extensions_len),
        'signature_alg_hash_hex': lambda self:
            to_hex(self.signature_algorithm_hash),
        'signature_alg_sign_hex': lambda self:
            to_hex(self.signature_algorithm_sign),
        'signature_b64': lambda self: encode_to_b64(self.signature),  # str

        'b64': lambda self: encode_to_b64(self.tdf)  # str
    }
)


def _parse_signature_input(tdf):
    with TdfBytesParser(tdf) as parser:
        parser.delegate('sct_version', _parse_version)
        parser.delegate('signature_type', _parse_signature_type)

        # rest of the SignatureInput is identical to an TimestampedEntry
        parser.delegate('_tmp', _parse_timestamped_entry)
        parser.res.update(parser.res['_tmp'])
        del parser.res['_tmp']

        return parser.result()


# 'digitally-signed struct' of the SignedCertificateTimestamp
SignatureInput = namedtuple(
    typename='SignatureInput',
    field_names='arg',
    lazy_vals={
        '_parse_func': lambda _: _parse_signature_input,

        'sct_version': lambda self: Version(self._parse['sct_version']),
        'signature_type': lambda self:
            SignatureType(self._parse['signature_type']),
        'timestamp': lambda self: int(self._parse['timestamp']),
        'entry_type': lambda self: LogEntryType(self._parse['entry_type']),
        'signed_entry': lambda self:
            ASN1Cert(self._parse['signed_entry'])
                if self.entry_type.is_x509_entry else
            PreCert(self._parse['signed_entry'])
                if self.entry_type.is_precert_entry else
            None,

        'precert_entry': lambda self: self._parse.get('precert_entry', None),
        'x509_entry': lambda self: self._parse.get('x509_entry', None),
    }
)


# 3.4 Merkle Tree
# https://tools.ietf.org/html/rfc6962#section-3.4


def _parse_merkle_leaf_type(tdf):
    with TdfBytesParser(tdf) as parser:
        parser.read('val', '!B')  # (255)
        return parser.result()


MerkleLeafType = namedtuple(
    typename='MerkleLeafType',
    field_names='arg',
    lazy_vals={
        '_parse_func': lambda _: _parse_merkle_leaf_type,

        'val': lambda self: int(self._parse['val']),

        'is_timestamped_entry': lambda self: self.val == 0,

        '__str__': lambda self: lambda:
            'timestamped_entry' if self.is_timestamped_entry else
            '<unkown merkle leaf type>',
    }
)


def _parse_timestamped_entry(tdf):
    with TdfBytesParser(tdf) as parser:
        parser.read('timestamp', '!Q')  # uint64  -> 8 bytes
        entry_type = LogEntryType(
            parser.delegate('entry_type', _parse_log_entry_type))

        # parse leaf_entry
        if entry_type.is_x509_entry:
            parser.delegate('signed_entry', _parse_asn1_cert)
            parser.res['x509_entry'] = parser.res['signed_entry']
        elif entry_type.is_precert_entry:
            parser.delegate('signed_entry', _parse_pre_cert)
            parser.res['precert_entry'] = parser.res['signed_entry']
        else:
            raise Exception(flo('Unknown entry_type number: {entry_type}'))

        # TODO DEBUG ctlog_get_entries.py related (it looks like some log
        #                                          answers are missing
        #                                          the ct_extensions,
        #                                         or an error in parse routines)
        try:
            parser.delegate('extensions', _parse_ct_extensions)
        except struct.error:
            pass

        return parser.result()


TimestampedEntry = namedtuple(
    typename='TimestampedEntry',
    field_names='arg',
    lazy_vals={
        '_parse_func': lambda _: _parse_timestamped_entry,

        'timestamp': lambda self: int(self._parse.get('timestamp')),
        'entry_type': lambda self: LogEntryType(self._parse['entry_type']),
        'signed_entry': lambda self:
            ASN1Cert(self._parse['signed_entry'])
                if self.entry_type.is_x509_entry else
            PreCert(self._parse['signed_entry'])
                if self.entry_type.is_precert_entry else
            None,
        'extensions': lambda self: CtExtensions(self._parse.get('extensions')),

        'precert_entry': lambda self: self._parse.get('precert_entry', None),
        'x509_entry': lambda self: self._parse.get('x509_entry', None),
    }
)


def _parse_merkle_tree_leaf(tdf):
    with TdfBytesParser(tdf) as parser:
        parser.delegate('version', _parse_version)
        leaf_type = parser.delegate('leaf_type', _parse_merkle_leaf_type)

        if MerkleLeafType(leaf_type).is_timestamped_entry:
            parser.delegate('leaf_entry', _parse_timestamped_entry)
        else:
            raise Exception('unknown leaf_type: {leaf_type}!')

        return parser.result()


MerkleTreeLeaf = namedtuple(
    typename='MerkleTreeLeaf',
    field_names='arg',
    lazy_vals={
        '_parse_func': lambda _: _parse_merkle_tree_leaf,

        'version': lambda self: Version(self._parse['version']),
        'leaf_type': lambda self: MerkleLeafType(self._parse['leaf_type']),
        'leaf_entry': lambda self: TimestampedEntry(self._parse['leaf_entry']),

        # alias for 'leaf_entry'
        'timestamped_entry': lambda self: self.leaf_entry,

        '__str__': lambda self: lambda:
            self.__repr__(),
    }
)


# 4.6. Retrieve Entries from Log
# https://tools.ietf.org/html/rfc6962#section-4.6

GetEntriesInput = namedtuple_utlz(
    typename='GetEntriesInput',
    field_names=[
        'start',
        'end',
    ],
)


GetEntriesResponseEntry = namedtuple_utlz(
    typename='GetEntriesResponseEntry',
    field_names=[
        'json_dict',
    ],
    lazy_vals={
        # The base-64 encoded MerkleTreeLeaf structure
        'leaf_input_b64': lambda self: self.json_dict['leaf_input'],
        'leaf_input_tdf': lambda self: decode_from_b64(self.leaf_input_b64),
        'leaf_input': lambda self: MerkleTreeLeaf(self.leaf_input_tdf),

        'is_x509_chain_entry': lambda self:
            self.leaf_input.timestamped_entry.entry_type == 0,
        'is_precert_chain_entry': lambda self: not self.is_x509_chain_entry,

        # The base-64 encoded unsigned data pertaining to the log entry.  In the
        # case of an X509ChainEntry, this is the "certificate_chain".  In the
        # case of a PrecertChainEntry, this is the whole "PrecertChainEntry"
        'extra_data_b64': lambda self: self.json_dict['extra_data'],
        'extra_data_tdf': lambda self: decode_from_b64(self.extra_data_b64),
        'extra_data': lambda self:
            X509ChainEntry(self.extra_data_tdf) if self.is_x509_chain_entry else
            PrecertChainEntry(self.extra_data_tdf),

        # '__str__': lambda self: '<GetEntriesResponseEntry>',
    }
)


GetEntriesResponse = namedtuple_utlz(
    typename='GetEntriesResponse',
    field_names=[
        'json_dict',
    ],
    lazy_vals={
        'entries': lambda self: [GetEntriesResponseEntry(entry)
                                 for entry
                                 in self.json_dict['entries']],

        # for convenience
        'first_entry': lambda self: self.entries[0],
    },
)



'''Decompose an ASN.1 certificate into its components tbsCertificate in DER
format, signatureAlgorithm in DER format, and signatureValue as bytes according
to https://tools.ietf.org/html/rfc5280#section-4.1
'''

import argparse
import base64

import pyasn1_modules.rfc5280
from pyasn1.codec.der.decoder import decode as der_decoder
from pyasn1.codec.der.encoder import encode as der_encoder
from utlz import flo

from ctutlz._version import __version__


def create_parser():
    '''Create the `ArgumentParser` for the command `decompose-cert`.'''
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('-v', '--version',
                        action='version',
                        default=False,
                        version=__version__,
                        help='print version number')

    req = parser.add_argument_group('required arguments')
    req.add_argument('--cert',
                     metavar='<filename>',
                     dest='cert_filename',
                     required=True,
                     help='Certificate in PEM, Base64, or DER format')

    parser.add_argument('--tbscert',
                        metavar='<filename>',
                        dest='tbscert_filename',
                        # required=True,
                        help='write extracted tbsCertificate to this file '
                             '(DER encoded)')
    parser.add_argument('--sign-algo',
                        metavar='<filename>',
                        dest='sign_algo_filename',
                        # required=True,
                        help='write extracted signatureAlgorithm to this file '
                             '(DER encoded)')
    parser.add_argument('--signature',
                        metavar='<filename>',
                        dest='sign_value_filename',
                        # required=True,
                        help='write extracted signatureValue to this file')
    return parser


def cert_der_from_data(cert_raw):
    '''Return DER encoded certificate from "raw" certificate data which could be
    encoded as PEM, Base64 (B64) or DER.

    Args:
        cert_raw(bytes):

    Return:
        DER encoded certificate as bytes
    '''
    try:
        # assume PEM or B64 format (str)
        cert_raw_str = cert_raw.decode('ascii')
        cert_b64 = cert_raw_str.split(
            '-----BEGIN CERTIFICATE-----', 1
        )[-1].split(
            '-----END CERTIFICATE-----'
        )[0].strip()
        cert_b64 = ''.join(cert_b64.splitlines())
        cert_der = base64.b64decode(cert_b64)
    except UnicodeDecodeError:
        # no PEM or B64 format; then, assume cert_raw is in DER format (bytes)
        cert_der = cert_raw
    return cert_der


# FIXME: put functionality to here when, refactoring to optionally read from
# stdin and write to stdout
def decompose():
    pass


def main():
    parser = create_parser()
    args = parser.parse_args()

    with open(args.cert_filename, 'rb') as fh:
        cert_raw = fh.read()
    cert_der = cert_der_from_data(cert_raw)
    cert, _ = der_decoder(cert_der,
                          asn1Spec=pyasn1_modules.rfc5280.Certificate())

    if args.tbscert_filename:
        tbscert_der = der_encoder(cert['tbsCertificate'])
        with open(args.tbscert_filename, 'wb') as fh:
            fh.write(tbscert_der)

    if args.sign_algo_filename:
        sign_algo_der = der_encoder(cert['signatureAlgorithm'])
        with open(args.sign_algo_filename, 'wb') as fh:
            fh.write(sign_algo_der)

    if args.sign_value_filename:
        signature_value = cert['signature'].asOctets()
        with open(args.sign_value_filename, 'wb') as fh:
            fh.write(signature_value)


if __name__ == '__main__':
    main()



import binascii
import socket
import struct
from functools import reduce

import certifi
import OpenSSL
import pyasn1_modules.rfc2560
import pyasn1_modules.rfc5280
from pyasn1.codec import ber
from pyasn1.codec.der.decoder import decode as der_decoder
from pyasn1.type.univ import ObjectIdentifier, OctetString, Sequence
from utlz import flo, namedtuple

from ctutlz.rfc6962 import SignedCertificateTimestamp
from ctutlz.sct.ee_cert import EndEntityCert, IssuerCert
from ctutlz.tls.sctlist import SignedCertificateTimestampList, TlsExtension18


def scts_from_cert(cert_der):
    '''Return list of SCTs of the SCTList SAN extension of the certificate.

    Args:
        cert_der(bytes): DER encoded ASN.1 Certificate

    Return:
        [<ctutlz.rfc6962.SignedCertificateTimestamp>, ...]
    '''
    cert, _ = der_decoder(
        cert_der, asn1Spec=pyasn1_modules.rfc5280.Certificate())
    sctlist_oid = ObjectIdentifier(value='1.3.6.1.4.1.11129.2.4.2')
    exts = []
    if 'extensions' in cert['tbsCertificate'].keys():
        exts = [extension
                for extension
                in cert['tbsCertificate']['extensions']
                if extension['extnID'] == sctlist_oid]

    if len(exts) != 0:
        extension_sctlist = exts[0]
        os_inner_der = extension_sctlist['extnValue']  # type: OctetString()
        os_inner, _ = der_decoder(os_inner_der, OctetString())
        sctlist_hex = os_inner.prettyPrint().split('0x')[-1]
        sctlist_der = binascii.unhexlify(sctlist_hex)

        sctlist = SignedCertificateTimestampList(sctlist_der)
        return [SignedCertificateTimestamp(entry.sct_der)
                for entry
                in sctlist.sct_list]
    return []


def sctlist_hex_from_ocsp_pretty_print(ocsp_resp):
    sctlist_hex = None
    splitted = ocsp_resp.split('<no-name>=1.3.6.1.4.1.11129.2.4.5', 1)
    if len(splitted) > 1:
        _, after = splitted
        _, sctlist_hex_with_rest = after.split('<no-name>=0x', 1)
        sctlist_hex, _ = sctlist_hex_with_rest.split('\n', 1)
    return sctlist_hex


def scts_from_ocsp_resp(ocsp_resp_der):
    '''Return list of SCTs of the OCSP status response.

    Args:
        ocsp_resp_der(bytes): DER encoded OCSP status response

    Return:
        [<ctutlz.rfc6962.SignedCertificateTimestamp>, ...]
    '''
    if ocsp_resp_der:
        ocsp_resp, _ = der_decoder(
            ocsp_resp_der, asn1Spec=pyasn1_modules.rfc2560.OCSPResponse())

        response_bytes = ocsp_resp.getComponentByName('responseBytes')
        if response_bytes is not None:
            # os: octet string
            response_os = response_bytes.getComponentByName('response')

            der_decoder.defaultErrorState = ber.decoder.stDumpRawValue
            response, _ = der_decoder(response_os, Sequence())

            sctlist_os_hex = sctlist_hex_from_ocsp_pretty_print(
                response.prettyPrint())

            if sctlist_os_hex:
                sctlist_os_der = binascii.unhexlify(sctlist_os_hex)
                sctlist_os, _ = der_decoder(sctlist_os_der, OctetString())
                sctlist_hex = sctlist_os.prettyPrint().split('0x')[-1]
                sctlist_der = binascii.unhexlify(sctlist_hex)

                sctlist = SignedCertificateTimestampList(sctlist_der)
                return [SignedCertificateTimestamp(entry.sct_der)
                        for entry
                        in sctlist.sct_list]
    return []


def scts_from_tls_ext_18(tls_ext_18_tdf):
    '''Return list of SCTs of the TLS extension 18 server reply.

    Args:
        tls_ext_18_tdf(bytes): TDF encoded TLS extension 18 server reply.

    Return:
        [<ctutlz.rfc6962.SignedCertificateTimestamp>, ...]
    '''
    scts = []

    if tls_ext_18_tdf:
        tls_extension_18 = TlsExtension18(tls_ext_18_tdf)
        sct_list = tls_extension_18.sct_list

        scts = [SignedCertificateTimestamp(entry.sct_der)
                for entry
                in sct_list]

    return scts


TlsHandshakeResult = namedtuple(
    typename='TlsHandshakeResult',
    field_names=[
        'ee_cert_der',      # (bytes)
        'issuer_cert_der',  # (bytes)
        'more_issuer_cert_der_candidates',  # [(bytes), ...]
        'ocsp_resp_der',    # (bytes)
        'tls_ext_18_tdf',   # (bytes)
        'err',              # (str)
    ],
    lazy_vals={
        'ee_cert': lambda self: EndEntityCert(self.ee_cert_der),
        'issuer_cert': lambda self: IssuerCert(self.issuer_cert_der),
        'more_issuer_cert_candidates': lambda self: [
            IssuerCert(cert_der)
            for cert_der
            in self.more_issuer_cert_der_candidates],

        'scts_by_cert': lambda self: scts_from_cert(self.ee_cert_der),
        'scts_by_ocsp': lambda self: scts_from_ocsp_resp(self.ocsp_resp_der),
        'scts_by_tls': lambda self: scts_from_tls_ext_18(self.tls_ext_18_tdf),
    }
)


def create_context(scts_tls, scts_ocsp, timeout):
    '''
    Args:
        scts_tls: If True, register callback for TSL extension 18 (for SCTs)
        scts_ocsp: If True, register callback for OCSP-response (for SCTs)
        timeout(int): timeout in seconds
    '''

    def verify_callback(conn, cert, errnum, depth, ok):
        '''Dummy callback.'''
        return 1  # True

    ctx = OpenSSL.SSL.Context(OpenSSL.SSL.SSLv23_METHOD)
    ctx.set_options(OpenSSL.SSL.OP_NO_SSLv2)
    ctx.set_options(OpenSSL.SSL.OP_NO_SSLv3)

    ctx.set_verify(OpenSSL.SSL.VERIFY_PEER, verify_callback)
    ca_filename = certifi.where()
    ctx.load_verify_locations(ca_filename)

    ctx.tls_ext_18_tdf = None
    if scts_tls:
        from ctutlz.tls.handshake_openssl import ffi, lib

        # this annotation makes the callback function available at
        # lib.serverinfo_cli_parse_cb() of type cdef so it can be used as
        # argument for the call of lib.SSL_CTX_add_client_custom_ext()
        @ffi.def_extern()
        def serverinfo_cli_parse_cb(ssl, ext_type, _in, inlen, al, arg):
            if ext_type == 18:

                def reduce_func(accum_value, current):
                    fmt = accum_value[0] + current[0]
                    values = accum_value[1] + (current[1], )
                    return fmt, values

                initializer = ('!', ())
                fmt, values = reduce(reduce_func, [
                    ('H', ext_type),
                    ('H', inlen),
                    (flo('{inlen}s'), bytes(ffi.buffer(_in, inlen))),
                ], initializer)
                ctx.tls_ext_18_tdf = struct.pack(fmt, *values)
            return 1  # True

        # register callback for TLS extension result into the SSL context
        # created with PyOpenSSL, using OpenSSL "directly"
        if not lib.SSL_CTX_add_client_custom_ext(ffi.cast('struct ssl_ctx_st *',
                                                          ctx._context),
                                                 18,
                                                 ffi.NULL, ffi.NULL, ffi.NULL,
                                                 lib.serverinfo_cli_parse_cb,
                                                 ffi.NULL):
            import sys
            sys.stderr.write('Unable to add custom extension 18\n')
            lib.ERR_print_errors_fp(sys.stderr)
            sys.exit(1)

    ctx.ocsp_resp_der = None
    if scts_ocsp:

        def ocsp_client_callback(connection, ocsp_data, data):
            ctx.ocsp_resp_der = ocsp_data
            return True

        ctx.set_ocsp_client_callback(ocsp_client_callback, data=None)

    ctx.set_timeout(timeout)

    return ctx


def create_socket(ctx):
    '''
    Args:
        ctx(OpenSSL.SSL.Context): OpenSSL context object
    '''
    raw_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    return OpenSSL.SSL.Connection(ctx, raw_sock)


def do_handshake(domain, scts_tls=True, scts_ocsp=True, timeout=5):
    '''
    Args:
        domain: string with domain name,
                for example: 'ritter.vg', or 'www.ritter.vg'
        scts_tls: If True, register callback for TSL extension 18 (for SCTs)
        scts_ocsp: If True, register callback for OCSP-response (for SCTs)
        timeout(int): timeout in seconds
    '''
    ctx = create_context(scts_tls, scts_ocsp, timeout)
    sock = create_socket(ctx)
    sock.request_ocsp()

    issuer_cert_x509 = None
    more_issuer_cert_candidates = []
    ee_cert_x509 = None
    ocsp_resp_der = None
    tls_ext_18_tdf = None
    err = ''

    try:
        sock.connect((domain, 443))
        sock.do_handshake()

        # type: OpenSSL.crypto.X509; ee: end entity
        ee_cert_x509 = sock.get_peer_certificate()
        # type: [OpenSSL.crypto.X509, ...]
        chain_x509s = sock.get_peer_cert_chain()
        if len(chain_x509s) > 1:
            issuer_cert_x509 = chain_x509s[1]
        more_issuer_cert_x509_candidates = [ee_cert_x509] + chain_x509s
        print(len(chain_x509s))

        ctx = sock.get_context()
        if scts_tls:
            if ctx.tls_ext_18_tdf:
                tls_ext_18_tdf = ctx.tls_ext_18_tdf
        if scts_ocsp:
            if ctx.ocsp_resp_der:
                ocsp_resp_der = ctx.ocsp_resp_der

    except Exception as exc:
        exc_str = str(exc)
        if exc_str == '':
            exc_str = str(type(exc))
        err = domain + ': ' + exc_str
    finally:
        sock.close()

    ee_cert_der = None
    if ee_cert_x509:
        ee_cert_der = OpenSSL.crypto.dump_certificate(
            type=OpenSSL.crypto.FILETYPE_ASN1,
            cert=ee_cert_x509)

    issuer_cert_der = None
    if issuer_cert_x509:
        # https://tools.ietf.org/html/rfc5246#section-7.4.2
        issuer_cert_der = OpenSSL.crypto.dump_certificate(
            type=OpenSSL.crypto.FILETYPE_ASN1,
            cert=issuer_cert_x509)

    more_issuer_cert_der_candidates = [
        OpenSSL.crypto.dump_certificate(type=OpenSSL.crypto.FILETYPE_ASN1,
                                        cert=cert_x509)
        for cert_x509
        in more_issuer_cert_x509_candidates
    ]

    return TlsHandshakeResult(ee_cert_der, issuer_cert_der,
                              more_issuer_cert_der_candidates,
                              ocsp_resp_der, tls_ext_18_tdf,
                              err)



import struct

from utlz import namedtuple as namedtuple_utlz


# tdf := "TLS Data Format" (cf. https://tools.ietf.org/html/rfc5246#section-4)


def namedtuple(typename, field_names='arg', lazy_vals=None, **kwargs):

    lazy_vals['_parse'] = lambda self: \
        self.arg if type(self.arg) == dict else \
        self._parse_func(self.arg)[0] if type(self.arg) == bytes else \
        None

    lazy_vals['tdf'] = lambda self: \
        self._parse['tdf']

    return namedtuple_utlz(typename, field_names, lazy_vals, **kwargs)


class TdfBytesParser(object):
    '''An instance of this is a file like object which enables access of a
    tdf (data) struct (a bytes string).
    '''

    # context methods

    def __init__(self, tdf_bytes):
        self._bytes = tdf_bytes
        self.offset = 0
        '''mutable parse results (read and delegate) dict'''
        self.res = {}

    def __enter__(self):
        self.offset = 0
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.offset = 0
        return

    # methods for parsing

    def read(self, key, fmt):
        data = struct.unpack_from(fmt, self._bytes, self.offset)
        self.offset += struct.calcsize(fmt)
        if len(data) == 1:
            self.res[key] = data[0]
        else:
            self.res[key] = data
        return self.res[key]

    def delegate(self, key, read_func):
        self.res[key], offset = read_func(self._bytes[self.offset:])
        self.offset += offset
        return self.res[key]

    def result(self):
        self.res['tdf'] = bytes(bytearray(self._bytes[0:self.offset]))
        return self.res, self.offset



import struct
from functools import reduce

from utlz import flo


def create_signature_input(ee_cert, sct, *_, **__):
    # cf. https://tools.ietf.org/html/rfc6962#section-3.2

    signature_type = 0  # 0 means certificate_timestamp
    entry_type = 0  # 0: ASN.1Cert, 1: PreCert

    def reduce_func(accum_value, current):
        fmt = accum_value[0] + current[0]
        values = accum_value[1] + (current[1], )
        return fmt, values

    initializer = ('!', ())

    # fmt = '!BBQh...', values = [<sct.version>, <signature_type>, ...]
    fmt, values = reduce(reduce_func, [
        ('B', sct.version.val),
        ('B', signature_type),
        ('Q', sct.timestamp),
        ('h', entry_type),

        # signed_entry
        ('B', ee_cert.len1),
        ('B', ee_cert.len2),
        ('B', ee_cert.len3),
        (flo('{ee_cert.len}s'), ee_cert.der),

        ('h', sct.extensions_len),
    ], initializer)
    return struct.pack(fmt, *values)


def create_signature_input_precert(ee_cert, sct, issuer_cert):
    # cf. https://tools.ietf.org/html/rfc6962#section-3.2

    signature_type = 0  # 0 means certificate_timestamp
    entry_type = 1  # 0: ASN.1Cert, 1: PreCert

    tbscert = ee_cert.tbscert.without_ct_extensions

    def reduce_func(accum_value, current):
        fmt = accum_value[0] + current[0]
        values = accum_value[1] + (current[1], )
        return fmt, values

    initializer = ('!', ())

    # fmt = '!BBQh...', values = [<sct.version>, <signature_type>, ...]
    fmt, values = reduce(reduce_func, [
        ('B', sct.version.val),
        ('B', signature_type),
        ('Q', sct.timestamp),
        ('h', entry_type),

        # signed_entry

        # issuer_key_hash[32]
        ('32s', issuer_cert.pubkey_hash),

        # tbs_certificate (rfc6962, page 12)
        #  * DER encoded TBSCertificate of the ee_cert
        #    * without SCT extension
        ('B', tbscert.len1),
        ('B', tbscert.len2),
        ('B', tbscert.len3),
        (flo('{tbscert.len}s'), tbscert.der),

        ('h', sct.extensions_len),
    ], initializer)
    return struct.pack(fmt, *values)



import struct

import OpenSSL
import pyasn1.error
from pyasn1_modules import rfc5280
from pyasn1.type.univ import ObjectIdentifier, Sequence
from pyasn1.codec.der.encoder import encode as der_encoder
from pyasn1.codec.der.decoder import decode as der_decoder

from utlz import namedtuple

from ctutlz.utils.string import string_without_prefix
from ctutlz.utils.encoding import sha256_digest


# https://hg.mozilla.org/mozilla-central/file/tip/security/certverifier/ExtendedValidation.cpp
EV_OIDs = ['1.2.392.200091.100.721.1',
           '1.3.6.1.4.1.6334.1.100.1',
           '2.16.756.1.89.1.2.1.1',
           '1.3.6.1.4.1.23223.1.1.1',
           '1.3.6.1.4.1.23223.1.1.1',
           '1.3.6.1.4.1.23223.1.1.1',
           '2.16.840.1.113733.1.7.23.6',
           '1.3.6.1.4.1.14370.1.6',
           '2.16.840.1.113733.1.7.48.1',
           '2.16.840.1.114404.1.1.2.4.1',
           '2.16.840.1.114404.1.1.2.4.1',
           '2.16.840.1.114404.1.1.2.4.1',
           '1.3.6.1.4.1.6449.1.2.1.5.1',
           '1.3.6.1.4.1.6449.1.2.1.5.1',
           '1.3.6.1.4.1.6449.1.2.1.5.1',
           '2.16.840.1.114413.1.7.23.3',
           '2.16.840.1.114413.1.7.23.3',
           '2.16.840.1.114414.1.7.23.3',
           '2.16.840.1.114414.1.7.23.3',
           '2.16.840.1.114412.2.1',
           '1.3.6.1.4.1.8024.0.2.100.1.2',
           '1.3.6.1.4.1.782.1.2.1.8.1',
           '2.16.840.1.114028.10.1.2',
           '1.3.6.1.4.1.4146.1.1',
           '1.3.6.1.4.1.4146.1.1',
           '1.3.6.1.4.1.4146.1.1',
           '2.16.578.1.26.1.3.3',
           '1.3.6.1.4.1.22234.2.5.2.3.1',
           '1.3.6.1.4.1.17326.10.14.2.1.2',
           '1.3.6.1.4.1.17326.10.8.12.1.2',
           '1.3.6.1.4.1.34697.2.1',
           '1.3.6.1.4.1.34697.2.2',
           '1.3.6.1.4.1.34697.2.3',
           '1.3.6.1.4.1.34697.2.4',
           '1.2.616.1.113527.2.5.1.1',
           '1.2.616.1.113527.2.5.1.1',
           '1.3.6.1.4.1.14777.6.1.1',
           '1.3.6.1.4.1.14777.6.1.2',
           '1.3.6.1.4.1.7879.13.24.1',
           '1.3.6.1.4.1.40869.1.1.22.3',
           '1.3.6.1.4.1.4788.2.202.1',
           '2.16.840.1.113733.1.7.23.6',
           '1.3.6.1.4.1.14370.1.6',
           '2.16.840.1.113733.1.7.48.1',
           '1.3.6.1.4.1.13177.10.1.3.10',
           '1.3.6.1.4.1.40869.1.1.22.3',
           '2.16.792.3.0.4.1.1.4',
           '1.3.159.1.17.1',
           '1.3.6.1.4.1.36305.2',
           '1.3.6.1.4.1.36305.2',
           '2.16.840.1.114412.2.1',
           '2.16.840.1.114412.2.1',
           '2.16.840.1.114412.2.1',
           '2.16.840.1.114412.2.1',
           '2.16.840.1.114412.2.1',
           '1.3.6.1.4.1.8024.0.2.100.1.2',
           '1.3.6.1.4.1.6449.1.2.1.5.1',
           '1.3.6.1.4.1.6449.1.2.1.5.1',
           '1.3.6.1.4.1.6449.1.2.1.5.1',
           '1.3.6.1.4.1.4146.1.1',
           '2.16.840.1.114028.10.1.2',
           '2.16.528.1.1003.1.2.7',
           '2.16.840.1.114028.10.1.2',
           '2.16.840.1.114028.10.1.2',
           '2.16.156.112554.3',
           '1.3.6.1.4.1.36305.2',
           '1.3.6.1.4.1.36305.2',
           '1.2.392.200091.100.721.1',
           '2.16.756.5.14.7.4.8',
           '1.3.6.1.4.1.22234.3.5.3.1',
           '1.3.6.1.4.1.22234.3.5.3.2',
           '1.3.6.1.4.1.22234.2.14.3.11',
           '1.3.6.1.4.1.22234.2.14.3.11',
           '1.3.6.1.4.1.22234.2.14.3.11',
           '2.16.840.1.113733.1.7.23.6',
           '2.23.140.1.1',
           '2.23.140.1.1',
           '2.23.140.1.1',
           '2.23.140.1.1',
           '2.23.140.1.1',
           '1.3.171.1.1.10.5.2']


def is_ev_cert(ee_cert):
    '''Return True if ee_cert is an extended validation certificate, else False.

    Args:
        ee_cert (EndEntityCert)
    '''
    oids = []
    oid_certificate_policies = ObjectIdentifier('2.5.29.32')

    all_extensions = ee_cert.tbscert.pyasn1['extensions']
    if all_extensions is not None:
        policy_extensions = [ext
                             for ext
                             in all_extensions
                             if ext['extnID'] == oid_certificate_policies]
        if len(policy_extensions) > 0:
            policy_extension = policy_extensions[0]
            sequence_der = policy_extension['extnValue']  # type: Sequence()
            try:
                sequence, _ = der_decoder(sequence_der, Sequence())
            except pyasn1.error.PyAsn1Error:
                sequence = []  # invalid encoded certificate policy extension

            for idx in range(len(sequence)):
                inner_sequence = sequence.getComponentByPosition(idx)
                oid = inner_sequence.getComponentByPosition(0)
                oids.append(str(oid))

    intersection = list(set(oids) & set(EV_OIDs))
    return intersection != []


def is_letsencrypt_cert(ee_cert):
    '''Return True if ee_cert was issued by Let's Encrypt.

    Args:
        ee_cert (EndEntityCert)
    '''
    organization_name_oid = ObjectIdentifier(value='2.5.4.10')
    issuer = ee_cert.tbscert.pyasn1['issuer']
    if issuer:
        for rdn in issuer['rdnSequence']:
            for item in rdn:
                if item.getComponentByName('type') == organization_name_oid:
                    organisation_name = str(item.getComponentByName('value'))
                    organisation_name = string_without_prefix('\x13\r',
                                                              organisation_name)
                    if organisation_name == "Let's Encrypt":
                        return True
    return False


def pyasn1_certificate_from_der(cert_der):
    '''Return pyasn1_modules.rfc5280.Certificate instance parsed from cert_der.
    '''
    cert, _ = der_decoder(cert_der, asn1Spec=rfc5280.Certificate())
    return cert


def copy_pyasn1_instance(instance):
    der = der_encoder(instance)
    copy, _ = der_decoder(der, rfc5280.TBSCertificate())
    return copy


def pyopenssl_certificate_from_der(cert_der):
    '''Return OpenSSL.crypto.X509 instance parsed from cert_der.
    '''
    cert = OpenSSL.crypto.load_certificate(type=OpenSSL.crypto.FILETYPE_ASN1,
                                           buffer=cert_der)
    return cert


def tbscert_without_sctlist(tbscert):
    '''Return pyasn1_modules.rfc2580.TBSCertificate instance `cert_pyasn1`
    without sctlist extension (OID 1.3.6.1.4.1.11129.2.4.2).
    '''
    sctlist_oid = ObjectIdentifier(value='1.3.6.1.4.1.11129.2.4.2')
    extensions = tbscert['extensions']
    without_sctlist = extensions.subtype()
    for extension in extensions:
        if extension['extnID'] != sctlist_oid:
            without_sctlist.append(extension)
    copy = copy_pyasn1_instance(tbscert)
    copy['extensions'] = without_sctlist
    return copy


def tbscert_without_ct_extensions(tbscert):
    '''Return pyasn1_modules.rfc5280.TBSCertificate instance `cert_pyasn1`
    without sctlist extension (OID 1.3.6.1.4.1.11129.2.4.3) and
    poison extension (OID 1.3.6.1.4.1.11129.2.4.2), if any.
    '''
    sctlist_oid = ObjectIdentifier(value='1.3.6.1.4.1.11129.2.4.2')
    poison_oid = ObjectIdentifier(value='1.3.6.1.4.1.11129.2.4.3')
    ct_oids = [sctlist_oid, poison_oid]

    extensions = tbscert['extensions']
    without_ct_extensions = extensions.subtype()
    for extension in extensions:
        if extension['extnID'] not in ct_oids:
            without_ct_extensions.append(extension)
    copy = copy_pyasn1_instance(tbscert)
    copy['extensions'] = without_ct_extensions
    return copy


TbsCert = namedtuple(
    typename='TbsCert',
    field_names=[
        'pyasn1',
    ],
    lazy_vals={
        'der': lambda self: der_encoder(self.pyasn1),
        'len': lambda self: len(self.der),
        'lens': lambda self: struct.unpack('!4B', struct.pack('!I', self.len)),
        # cf. https://tools.ietf.org/html/rfc6962#section-3.2      <1..2^24-1>
        'len1': lambda self: self.lens[1],
        'len2': lambda self: self.lens[2],
        'len3': lambda self: self.lens[3],

        'without_ct_extensions':
            lambda self: TbsCert(tbscert_without_ct_extensions(self.pyasn1)),
    }
)


EndEntityCert = namedtuple(
    typename='EndEntityCert',
    field_names=[
        'der',
        'issuer_cert=None',  # type: der
    ],
    lazy_vals={
        'len': lambda self: len(self.der),
        # cf. https://tools.ietf.org/html/rfc6962#section-3.2      <1..2^24-1>
        'lens': lambda self: struct.unpack('!4B', struct.pack('!I', self.len)),
        'len1': lambda self: self.lens[1],
        'len2': lambda self: self.lens[2],
        'len3': lambda self: self.lens[3],

        'pyasn1': lambda self: pyasn1_certificate_from_der(self.der),
        'tbscert': lambda self: TbsCert(self.pyasn1['tbsCertificate']),

        # FIXME: YAGNI?
        'pyopenssl': lambda self: pyopenssl_certificate_from_der(self.der),

        'is_ev_cert': lambda self: is_ev_cert(self),
        'is_letsencrypt_cert': lambda self: is_letsencrypt_cert(self),
    }
)


IssuerCert = namedtuple(
    typename='IssuerCert',
    field_names=[
        'der',
    ],
    lazy_vals={
        'pyasn1': lambda self: pyasn1_certificate_from_der(self.der),
        'pubkey_pyasn1': lambda self:
            self.pyasn1['tbsCertificate']['subjectPublicKeyInfo'],
        'pubkey_der': lambda self: der_encoder(self.pubkey_pyasn1),
        'pubkey_hash': lambda self: sha256_digest(self.pubkey_der),
    }
)



from __future__ import unicode_literals  # for Python-2
import json
import re
try:
    # since python 3.6
    import urllib.request as urllib_request
except ImportError:
    import urllib as urllib_request
from os.path import abspath, expanduser, join, isfile, dirname

import html2text
from utlz import load_json, namedtuple, text_with_newlines
from utlz.types import Enum

from ctutlz.utils.encoding import decode_from_b64, encode_to_b64
from ctutlz.utils.encoding import digest_from_b64
from ctutlz.utils.logger import logger


# states 'compliant with chrome ct policy'
ChromeStates = Enum(INCLUDED='included',
                    FROZEN='frozen',

                    PENDING='pending for inclusion',

                    DISQUALIFIED='disqualified',
                    REJECTED='rejected',
                    DISTRUSTED='distrusted')

Log = namedtuple(
    typename='Log',
    field_names=[  # each of type: str
        'description',
        'key',     # base-64 encoded, type: str
        'url',
        'maximum_merge_delay',
        'operated_by',

        # optional infos from *.json
        'final_sth=None',
        'disqualified_at=None',
        'dns_api_endpoint=None',

        # optional infos from webpage 'known logs'
        'contact=None',
        'chrome_bug=None',
        'notes=None',
        'id_b64_non_calculated=None',
        'certificate_expiry_range=None',

        # ChromeStates attribute or None
        'chrome_state=None',

        # TODO add real checks; interact with logs (-> ctlog.py):

        # True, False, or None
        'active=None',        # log accepts new entries
        # True, False, or None
        'alive=None',         # log answers
    ],
    lazy_vals={
        'key_der': lambda self: decode_from_b64(self.key),  # type: bytes
        'id_der': lambda self: digest_from_b64(self.key),   # type: bytes
        'id_b64': lambda self: encode_to_b64(self.id_der),  # type: str
        'pubkey': lambda self: '\n'.join([                  # type: str
                                          '-----BEGIN PUBLIC KEY-----',
                                          text_with_newlines(text=self.key,
                                                             line_length=64),
                                          '-----END PUBLIC KEY-----']),
        'scts_accepted_by_chrome':
            lambda self:
                None if self.chrome_state is None else
                True if self.chrome_state in [ChromeStates.INCLUDED,
                                              ChromeStates.FROZEN] else
                False,
    }
)


# plurale tantum constructor
def Logs(log_dicts):
    '''
    Arg log_dicts example: [
        {
            'description': ...,
            'key': ...,
            'url': ...,
            'maximum_merge_delay': ...,
            'operated_by': ...,

            'final_sth': ...,  # optional
            'disqualified_at=None': ...,
            'dns_api_endpoint=None': ...,

            # optional
            'started': ...,
            'submitted_for_inclusion_in_chrome': ...,
            'contact': ...,
            'chrome_bug'. ...,
            'notes': ...,
            'id_b64_non_calculated': ...,

            'chrome_state', CHROME_STATE.<enum>
        },
    ]
    '''
    for log in log_dicts:
        log.pop('id_b64', None)
        log.pop('pubkey', None)
        log.pop('scts_accepted_by_chrome', None)
    return [Log(**kwargs) for kwargs in log_dicts]


def set_operator_names(logs_dict):
    '''For each log in logs_dict overwrite the operator numbers by operator
    names.

    logs_dict example: {
      'logs: [
        {
          "description": "Google 'Aviator' log",
          "key": "MFkwE..."
          "url": "ct.googleapis.com/aviator/",
          "maximum_merge_delay": 86400,
          "operated_by": [0],
          "final_sth": {
            ...
          },
          "dns_api_endpoint": ...
        },
      ],
      'operators': [
        ...
      ]
    }
    '''
    for log in logs_dict['logs']:
        operator_ids = log['operated_by']
        operator_names = [operator['name']
                          for operator
                          in logs_dict['operators']
                          if operator['id'] in operator_ids]
        log['operated_by'] = operator_names


def unset_operator_names(logs_dict):
    '''For each log in logs_dict reset the operator names by operator numbers.

    logs_dict example: {
      'logs: [
        {
          "description": "Google 'Aviator' log",
          "key": "MFkwE..."
          "url": "ct.googleapis.com/aviator/",
          "maximum_merge_delay": 86400,
          "operated_by": ["Google"],
          "final_sth": {
            ...
          },
          "dns_api_endpoint": ...
        },
      ],
      'operators': [
        ...
      ]
    }
    '''
    for log in logs_dict['logs']:
        operator_names = log['operated_by']

        operator_ids = [operator['id']
                        for operator
                        in logs_dict['operators']
                        if operator['name'] in operator_names]
        log['operated_by'] = operator_ids


'''logs included in chrome browser'''
BASE_URL = 'https://www.gstatic.com/ct/log_list/'
URL_LOG_LIST = BASE_URL + 'log_list.json'
URL_ALL_LOGS = BASE_URL + 'all_logs_list.json'


def download_log_list(url=URL_ALL_LOGS):
    '''Download json file with known logs accepted by chrome and return the
    logs as a list of `Log` items.

    Return: dict, the 'logs_dict'

    logs_dict example: {
      'logs: [
        {
          "description": "Google 'Aviator' log",
          "key": "MFkwE..."
          "url": "ct.googleapis.com/aviator/",
          "maximum_merge_delay": 86400,
          "operated_by": [0],
          "final_sth": {
            ...
          },
          "dns_api_endpoint": ...
        },
      ],
      'operators': [
        ...
      ]
    }
    '''
    response = urllib_request.urlopen(url)
    response_str = response.read()
    try:
        data = json.loads(response_str)
    except TypeError:
        # Python-3.x < 3.6
        data = json.loads(response_str.decode('utf-8'))
    return data


def read_log_list(filename):
    '''Read log list from file `filename` and return as logs_dict.

    Return: dict, the 'logs_dict'

    logs_dict example: {
      'logs: [
        {
          "description": "Google 'Aviator' log",
          "key": "MFkwE..."
          "url": "ct.googleapis.com/aviator/",
          "maximum_merge_delay": 86400,
          "operated_by": [0],
          "final_sth": {
            ...
          },
          "dns_api_endpoint": ...
        },
      ],
      'operators': [
        ...
      ]
    }
    '''
    filename = abspath(expanduser(filename))
    data = load_json(filename)
    return data


def get_log_list(list_name='really_all_logs.json'):
    '''Try to read log list from local file.  If file not exists download
    log list.

    Return: dict, the 'logs_dict'

    logs_dict example: {
      'logs: [
        {
          "description": "Google 'Aviator' log",
          "key": "MFkwE..."
          "url": "ct.googleapis.com/aviator/",
          "maximum_merge_delay": 86400,
          "operated_by": [0],
          "final_sth": {
            ...
          },
          "dns_api_endpoint": ...
        },
      ],
      'operators': [
        ...
      ]
    }
    '''
    thisdir = dirname(__file__)
    filename = join(thisdir, list_name)
    if isfile(filename):
        logs_dict = read_log_list(filename)
    else:
        logs_dict = download_log_list(''.join([BASE_URL, list_name]))
    return logs_dict


def _log_dict_from_log_text(log_text):
    '''
    Args:
        log_text example (str):
            """
            ct.googleapis.com/pilot

            Base64 Log ID: pLkJkLQYWBSHuxOizGdwCjw1mAT5G9+443fNDsgN3BA=
            Operator: Google
            Started: 2013-03-25
            HTTPS supported: yes

            Maximum Merge Delay: 24 hours
            Contact: google-ct-logs@googlegroups.com
            Chrome inclusion status: Included."""

    return example: {
        'url': 'ct.googleapis.com/pilot/',
        'id_b64_non_calculated': 'pLkJkLQYWBSHuxOizGdwCj'
                                 'w1mAT5G9+443fNDsgN3BA=',
        'operated_by': ['Google'],
        'started': '2013-03-25',
        'https_supported': 'yes',
        'maximum_merge_delay': 86400,
        'contact': 'google-ct-logs@googlegroups.com',
        'chrome_inclusion_status': 'Included.',
        'chrome_status': 'included',
        'description': None,
        'key': None,
    }
    '''
    res = re.match(r'''(?P<log_url>[^ \n]+)
                       (?P<key_vals>(:?[^:]+:.+[\n$])*)
                       (?P<notes>(.*[\n$])*)''',
                   log_text,
                   flags=re.M | re.X)
    res_dict = res.groupdict()

    log_dict = {
        # normalize: assure url ends with a '/'
        'url': res_dict['log_url'].rstrip('/') + '/',
    }

    notes = res_dict.get('notes', '').strip()
    if notes:
        notes = re.sub(r'\s+', ' ', notes)  # just a oneliner, no formattings
        log_dict['notes'] = notes

    key_vals = res_dict.get('key_vals', '')
    for line in key_vals.split('\n'):
        if ':' in line:
            key, val = line.split(':', 1)
            val = val.strip()
            key = key.strip().lower().replace(' ', '_')
            if key == 'mmd':
                key = 'maximum_merge_delay'
            if key == 'maximum_merge_delay':
                if val.endswith('hours'):
                    hours = val.split(' ')[0]
                    val = int(hours) * 3600
            if key == 'base64_log_id':
                key = 'id_b64_non_calculated'
            if key == 'operator':
                key = 'operated_by'
                val = [val]
            log_dict[key] = val

    for key in 'description', 'key', 'maximum_merge_delay':
        # set to default `None` if they not exist
        log_dict[key] = log_dict.get(key, None)

    return log_dict


def _text_from_html(html):
    h2t = html2text.HTML2Text()
    h2t.ignore_links = True
    return h2t.handle(html)


def _logs_dict_from_html(html):
    try:
        text = _text_from_html(html)  # Python-3
    except UnicodeDecodeError:
        text = _text_from_html(html.decode('utf-8'))  # Python-2

    # ged rid of ending section
    text = text.split('###  Known Logs')[-1]

    # FIXME not needed anymore after redesign of known-logs.html?
    try:
        # remove erroneous ####-headers
        text = re.sub(r'####\s*$', '', text, flags=re.M)
        text = re.sub(r'####\s+(?=.*: )', '', text, flags=re.M)
        # remove backticks (from surrounding log ids)
        text = re.sub(r'`', '', text, flags=re.M)
    except Exception:  # InvocationError:
        # Python-2.6
        # remove erroneous ####-headers
        text = re.sub(r'####\s*$', '', text)
        text = re.sub(r'####\s+(?=.*: )', '', text)
        # remove backticks (from surrounding log ids)
        # (does not work at Python-2.6: text = re.sub(r'`', '', text, re.M) )
        text = text.replace('`', '')

    # normalize spaces (for Python-2)
    text = text.replace('\xa0', ' ')
    # remove trailing spaces
    text = '\n'.join([line.rstrip() for line in text.split('\n')])

    text = text.split('\n\n---\n')[0]  # get rid of bottom comments section

    # logger.debug(text)

    # keep only text blocks with log lists, drop first section
    log_blocks = text.split('\n### ', -1)[1:]

    # logger.debug(log_blocks)

    logs_dict = {}
    for text_block in log_blocks:
        # eg. (' Active Logs', ' \n\n#### ct.googleapis.com/pilot\n...')
        name, rest = text_block.split('\n\n', 1)
        # title of the log block,  eg. title = 'included in chrome'
        title = name.strip().lower().replace(' ', '_')

        # FIXME: simplify. Currently only special-purpose-logs are listed on
        #        known-log.html
        chrome_state = None
        if title.startswith('included'):
            chrome_state = ChromeStates.INCLUDED
        elif title.startswith('frozen'):
            chrome_state = ChromeStates.FROZEN
        elif title.startswith('pending'):
            chrome_state = ChromeStates.PENDING
        elif title.startswith('disqualified'):
            chrome_state = ChromeStates.DISQUALIFIED
        elif title.startswith('rejected'):
            chrome_state = ChromeStates.REJECTED
        elif 'distrusted' in title:
            chrome_state = ChromeStates.DISTRUSTED
        elif title.startswith('other'):
            chrome_state = None
        elif title.startswith('special_purpose_logs'):
            chrome_state = None
        else:
            raise Exception('unknown chrome_state for log-text_block')

        logs_dict[title] = []
        for log_text in rest.strip().lstrip('#### ').split('\n\n#### '):
            if log_text.strip() != '':
                log_dict = _log_dict_from_log_text(log_text)
                log_dict['chrome_state'] = chrome_state

                logs_dict[title].append(log_dict)

    return logs_dict


def logs_dict_from_webpage():
    '''Download log-list webpage and return as logs_dict.'''
    url = 'https://www.certificate-transparency.org/known-logs'
    response = urllib_request.urlopen(url)
    html = response.read().decode('utf-8')
    return _logs_dict_from_html(html)


def print_schema():
    thisdir = dirname(__file__)
    filename = join(thisdir, 'log_list_schema.json')
    with open(filename, 'r') as fh:
        json_str = fh.read()
    # print(json_str.strip())
    print('TODO')



def to_hex(val):
    '''Return val as str of hex values concatenated by colons.'''
    if type(val) is int:
        return hex(val)
    try:
        # Python-2.x
        if type(val) is long:
            return hex(val)
    except NameError:
        pass
    # else:
    try:
        # Python-2.x
        return ":".join("{0:02x}".format(ord(char)) for char in val)
    except TypeError:
        # Python-3.x
        return ":".join("{0:02x}".format(char) for char in val)


# http://stackoverflow.com/a/16891418
def string_without_prefix(prefix ,string):
    '''Return string without prefix.  If string does not start with prefix,
    return string.
    '''
    if string.startswith(prefix):
        return string[len(prefix):]
    return string


def string_with_prefix(prefix, string):
    '''Return string with prefix prepended.  If string already starts with
    prefix, return string.
    '''
    return str(prefix) + string_without_prefix(str(prefix), str(string))



import logging
import sys

logger = logging.getLogger('ctutlz')

VERBOSE = 15  # between DEBUG=10 and INFO=20


class InfoFilter(logging.Filter):
    def filter(self, rec):
        return rec.levelno in (logging.DEBUG, VERBOSE, logging.INFO)


def init_logger():
    logging.addLevelName(VERBOSE, "VERBOSE")

    def info_verbose(self, message, *args, **kws):
        self.log(VERBOSE, message, *args, **kws)

    logging.Logger.verbose = info_verbose


def setup_logging(loglevel):
    '''Write info, verbose and debug messages to stdout, else to stderr
    (warning and error).
    '''
    logger.setLevel(loglevel)
    try:
        # python 2.6
        out_handler = logging.StreamHandler(stream=sys.stdout)
        err_handler = logging.StreamHandler(stream=sys.stderr)
    except TypeError:
        # since python 2.7
        out_handler = logging.StreamHandler()
        err_handler = logging.StreamHandler(stream=sys.stderr)

    out_handler.setLevel(loglevel)
    out_handler.addFilter(InfoFilter())

    err_handler.setLevel(logging.WARNING)

    logger.addHandler(out_handler)
    logger.addHandler(err_handler)

    return logger



'''Submit a certificate chain to a Certificate Transparency log (CT-log);
received SCTs will be validated.
'''

import argparse
import logging

from utlz import flo, first_paragraph

from ctutlz.utils.logger import setup_logging, VERBOSE


def create_parser():
    parser = argparse.ArgumentParser(description=first_paragraph(__doc__))
    parser.add_argument('url',
                        nargs='+',
                        help='CT-log url '
                        '(example: `ct.googleapis.com/pilot`; cf. '
                        'https://www.certificate-transparency.org/known-logs)')

    meg = parser.add_mutually_exclusive_group()
    meg.add_argument('--short',
                     dest='loglevel',
                     action='store_const',
                     const=logging.INFO,
                     default=VERBOSE,  # default loglevel if nothing set
                     help='show short result and warnings/errors only')
    meg.add_argument('--debug',
                     dest='loglevel',
                     action='store_const',
                     const=logging.DEBUG,
                     help='show more for diagnostic purposes')

    req = parser.add_argument_group("exactly-one args "
                                    "(check the 'usage' above)")
    meg = req.add_mutually_exclusive_group(required=True)
    meg.add_argument('--chain',
                     metavar='<file>',
                     help='chain file with PEM encoded certificates '
                          '(end entity cert first; '
                          'with or without root ca cert)')
    meg.add_argument('--certs',
                     nargs='+',
                     metavar='<file>',
                     help='DER or PEM encoded certificates '
                          'which build the chain '
                          '(ee cert first; with or without root ca cert)')
    return parser


def main():
    parser = create_parser()
    args = parser.parse_args()
    logger = setup_logging(args.loglevel)
    logger.debug(args)

    for log_url in args.url:
        print(log_url)
        # TODO DEVEL


if __name__ == '__main__':
    main()



'''Verify Signed Certificate Timestamps (SCTs) delivered from one or several
hosts by X.509v3 extension, TLS extension, or OCSP stapling.

A lot of the functionality originally comes and have been learned from the
script sct-verify.py written by Pier Carlo Chiodi:
https://github.com/pierky/sct-verify/blob/master/sct-verify.py (under GPL)

He also described the SCT verification steps very well in his blog:
https://blog.pierky.com/certificate-transparency-manually-verify-sct-with-openssl/
'''

import argparse
import logging
import struct

from utlz import flo, first_paragraph, text_with_newlines

from ctutlz.tls.handshake import do_handshake
from ctutlz.ctlog import download_log_list, get_log_list, read_log_list
from ctutlz.ctlog import Logs, set_operator_names
from ctutlz.sct.verification import verify_scts
from ctutlz.sct.signature_input import create_signature_input_precert
from ctutlz.sct.signature_input import create_signature_input
from ctutlz.utils.string import to_hex
from ctutlz.utils.logger import VERBOSE, init_logger, setup_logging, logger
from ctutlz._version import __version__


def create_parser():
    parser = argparse.ArgumentParser(description=first_paragraph(__doc__))
    parser.add_argument('hostname',
                        nargs='+',
                        help="host name of the server (example: 'ritter.vg')")
    parser.add_argument('-v', '--version',
                        action='version',
                        default=False,
                        version=__version__,
                        help='print version number')

    meg = parser.add_mutually_exclusive_group()
    meg.add_argument('--short',
                     dest='loglevel',
                     action='store_const',
                     const=logging.INFO,
                     default=VERBOSE,  # default loglevel if nothing set
                     help='show short results and warnings/errors only')
    meg.add_argument('--debug',
                     dest='loglevel',
                     action='store_const',
                     const=logging.DEBUG,
                     help='show more for diagnostic purposes')

    meg1 = parser.add_mutually_exclusive_group()
    meg1.add_argument('--cert-only',
                      dest='verification_tasks',
                      action='store_const',
                      const=[verify_scts_by_cert],
                      default=[verify_scts_by_cert,
                               verify_scts_by_tls,
                               verify_scts_by_ocsp],
                      help='only verify SCTs included in the certificate')
    meg1.add_argument('--tls-only',
                      dest='verification_tasks',
                      action='store_const',
                      const=[verify_scts_by_tls],
                      help='only verify SCTs gathered from TLS handshake')
    meg1.add_argument('--ocsp-only',
                      dest='verification_tasks',
                      action='store_const',
                      const=[verify_scts_by_ocsp],
                      help='only verify SCTs gathered via OCSP request')

    meg2 = parser.add_mutually_exclusive_group()
    meg2.add_argument('--log-list',
                      dest='log_list_filename',
                      metavar='<filename>',
                      help='filename of a log list in JSON format')
    meg2.add_argument('--latest-logs',
                      dest='fetch_ctlogs',
                      action='store_const',
                      const=download_log_list,
                      default=get_log_list,
                      help='for SCT verification against known CT Logs '
                           "(compliant with Chrome's CT policy) "
                           'download latest version of '
                           'https://www.certificate-transparency.org/'
                           'known-logs/all_logs_list.json '
                           '-- use built-in log list really_all_logs.json '
                           'from 2017-08-11 if --latest-logs or --log-list '
                           'are not set')
    return parser


def verify_scts_by_cert(res, ctlogs):
    '''
    Args:
        res(ctutlz.tls.TlsHandshakeResult)
        ctlogs([<ctutlz.ctlog.Log>, ...])

    Return:
        [<ctutlz.sct.verification.SctVerificationResult>, ...]
    '''
    return verify_scts(
        ee_cert=res.ee_cert,
        scts=res.scts_by_cert,
        logs=ctlogs,
        issuer_cert=res.issuer_cert,
        more_issuer_cert_candidates=res.more_issuer_cert_candidates,
        sign_input_func=create_signature_input_precert)


def verify_scts_by_tls(res, ctlogs):
    '''
    Args:
        res(ctutlz.tls.TlsHandshakeResult)
        ctlogs([<ctutlz.ctlog.Log>, ...])

    Return:
        [<ctutlz.sct.verification.SctVerificationResult>, ...]
    '''
    return verify_scts(
        ee_cert=res.ee_cert,
        scts=res.scts_by_tls,
        logs=ctlogs,
        issuer_cert=None,
        more_issuer_cert_candidates=None,
        sign_input_func=create_signature_input)


def verify_scts_by_ocsp(res, ctlogs):
    '''
    Args:
        res(ctutlz.tls.TlsHandshakeResult)
        ctlogs([<ctutlz.ctlog.Log>, ...])

    Return:
        [<ctutlz.sct.verification.SctVerificationResult>, ...]
    '''
    return verify_scts(
        ee_cert=res.ee_cert,
        scts=res.scts_by_ocsp,
        logs=ctlogs,
        issuer_cert=None,
        more_issuer_cert_candidates=None,
        sign_input_func=create_signature_input)


# for more convenient command output
verify_scts_by_cert.__name__ = 'SCTs by Certificate'
verify_scts_by_tls.__name__ = 'SCTs by TLS'
verify_scts_by_ocsp.__name__ = 'SCTs by OCSP'


def show_signature_verbose(signature):
    '''Print out signature as hex string to logger.verbose.

    Args:
        signature(bytes)
    '''
    sig_offset = 0
    while sig_offset < len(signature):
        if len(signature) - sig_offset > 16:
            bytes_to_read = 16
        else:
            bytes_to_read = len(signature) - sig_offset
        sig_bytes = struct.unpack_from(flo('!{bytes_to_read}s'),
                                       signature,
                                       sig_offset)[0]
        if sig_offset == 0:
            logger.verbose('Signature : %s' % to_hex(sig_bytes))
        else:
            logger.verbose('            %s' % to_hex(sig_bytes))
        sig_offset = sig_offset + bytes_to_read


def show_verification(verification):
    '''
    Args:
        verification(ctutlz.sct.verification.SctVerificationResult)
    '''
    sct = verification.sct

    sct_log_id1, sct_log_id2 = [to_hex(val)
                                for val
                                in struct.unpack("!16s16s", sct.log_id.tdf)]
    logger.info('```')
    logger.verbose('=' * 59)
    logger.verbose(flo('Version   : {sct.version_hex}'))
    logger.verbose(flo('LogID     : {sct_log_id1}'))
    logger.verbose(flo('            {sct_log_id2}'))
    logger.info(flo('LogID b64 : {sct.log_id_b64}'))
    logger.verbose(flo('Timestamp : {sct.timestamp} ({sct.timestamp_hex})'))
    logger.verbose(flo(
        'Extensions: {sct.extensions_len} ({sct.extensions_len_hex})'))
    logger.verbose(flo('Algorithms: {sct.signature_alg_hash_hex}/'
                       '{sct.signature_algorithm_signature} (hash/sign)'))

    show_signature_verbose(sct.signature)
    prefix = 'Sign. b64 : '
    logger.info(prefix + text_with_newlines(sct.signature_b64, line_length=16*3,
                                            newline='\n' + ' '*len(prefix)))

    logger.verbose('--')  # visual gap between sct infos and verification result

    log = verification.log
    if log is None:
        logger.info('Log not found\n')
    else:
        logger.info(flo('Log found : {log.description}'))
        logger.verbose('Operator  : ' + ', '.join(log.operated_by))
        logger.info(flo('Chrome    : {log.chrome_state}'))

    if verification.verified:
        logger.info(flo('Result    : Verified OK'))
    else:
        logger.info(flo('Result    : Verification Failure'))

    logger.info('```\n')


def scrape_and_verify_scts(hostname, verification_tasks, ctlogs):
    logger.info(flo('# {hostname}\n'))

    res = do_handshake(hostname,
                       scts_tls=(verify_scts_by_tls in verification_tasks),
                       scts_ocsp=(verify_scts_by_ocsp in verification_tasks))
    if res.ee_cert_der:
        logger.debug('got certificate\n')
        if res.ee_cert.is_ev_cert:
            logger.info('* EV cert')
        else:
            logger.info('* no EV cert')
        if res.ee_cert.is_letsencrypt_cert:
            logger.info("* issued by Let's Encrypt\n")
        else:
            logger.info("* not issued by Let's Encrypt\n")

    if res.err:
        logger.warn(res.err)
    else:
        for verification_task in verification_tasks:
            logger.info(flo('## {verification_task.__name__}\n'))
            verifications = verification_task(res, ctlogs)
            if verifications:
                for verification in verifications:
                    show_verification(verification)
            elif res.ee_cert_der is not None:
                logger.info('no SCTs\n')


def main():
    init_logger()
    parser = create_parser()
    args = parser.parse_args()
    setup_logging(args.loglevel)
    logger.debug(args)

    # set ctlogs, type: [<ctutlz.ctlog.Log>, ...]
    logs_dict = args.fetch_ctlogs()
    set_operator_names(logs_dict)
    ctlogs = Logs(logs_dict['logs'])
    if args.log_list_filename:
        logs_dict = read_log_list(args.log_list_filename)
        set_operator_names(logs_dict)
        ctlogs = Logs(logs_dict['logs'])

    for host in args.hostname:
        scrape_and_verify_scts(host, args.verification_tasks, ctlogs)


if __name__ == '__main__':
    # when calling `verify-scts` directly from source as pointed out in the
    # README.md (section Devel-Commands) the c-code part needs to be compiled,
    # else the import of the c-module `ctutlz.tls.handshake_openssl` would fail.
    import ctutlz.tls.handshake_openssl_build
    ctutlz.tls.handshake_openssl_build.compile()

    main()



'''Retrieve an entry from a Certificate Transparency Log (ct-log),
cf. https://tools.ietf.org/html/rfc6962#section-4.6
'''

import argparse
import logging
import socket
import ssl
from pprint import pprint

import certifi
import requests
import OpenSSL.crypto
from hexdump import hexdump
from OpenSSL._util import lib as cryptolib
from utlz import flo, first_paragraph

from ctutlz import rfc6962
from ctutlz.sct.verification import verify_signature
from ctutlz.utils.encoding import decode_from_b64
from ctutlz.utils.logger import setup_logging
from ctutlz.utils.string import string_with_prefix, string_without_prefix


def create_parser():
    parser = argparse.ArgumentParser(description=first_paragraph(__doc__))
    meg = parser.add_mutually_exclusive_group()
    meg.add_argument('--short',
                     dest='loglevel',
                     action='store_const',
                     const=logging.WARNING,
                     default=logging.INFO,  # default loglevel if nothing set
                     help='show short result and warnings/errors only')
    meg.add_argument('--debug',
                     dest='loglevel',
                     action='store_const',
                     const=logging.DEBUG,
                     help='show more for diagnostic purposes')

    req = parser.add_argument_group('required arguments')
    req.add_argument('--log',
                     metavar='<url>',
                     required=True,
                     help='ct-log to retrieve '
                          '(example: `ct.googleapis.com/pilot`; cf. '
                          'https://www.certificate-transparency.org/'
                          'known-logs)')
    req.add_argument('--entry',
                     metavar='<dec>',
                     required=True,
                     help='zero-based decimal index of the log entry '
                          'to retrieve')
    req.add_argument('--signature',
                     metavar='<b64>',
                     required=True,
                     help='SCT signature, base-64 encoded')
    return parser


def get_peercert_pem(url):
    peercert_pem = None

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    ssl_sock = ssl.wrap_socket(sock,
                               cert_reqs=ssl.CERT_OPTIONAL,
                               ca_certs=certifi.where())
    ssl_sock.settimeout(30)
    try:
        ssl_sock.connect((url, 443))

        peercert_der = ssl_sock.getpeercert(binary_form=True)
        peercert_pem = ssl.DER_cert_to_PEM_cert(peercert_der)
    finally:
        ssl_sock.close()

    return peercert_pem


# https://stackoverflow.com/a/30929459
def pkey2pem(pkey):
    bio = OpenSSL.crypto._new_mem_buf()
    cryptolib.PEM_write_bio_PUBKEY(bio, pkey._pkey)
    return OpenSSL.crypto._bio_to_string(bio)


def get_pubkey_pem(url):
    '''Return pubkey from url's peercert PEM encoded as bytestring.'''
    peercert_pem = get_peercert_pem(url)
    x509 = OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_PEM,
                                           peercert_pem)
    pkey = x509.get_pubkey()
    pubkey_pem = pkey2pem(pkey)
    return pubkey_pem


def get_entry(index, ctlog_url, signature_b64):
    url = string_with_prefix('https://', ctlog_url) + '/ct/v1/get-entries'
    inp = rfc6962.GetEntriesInput(start=index, end=index)
    req = requests.get(url, params=inp._asdict())
    if req.ok:
        print('--')
        # from pprint import pprint
        # pprint(req.json())
        # print(len(req.json()))
        response = rfc6962.GetEntriesResponse(json_dict=req.json())
        # pprint(response.json_dict)
        print('--')
        leaf_input = response.first_entry.leaf_input
        # print(leaf_input.version)
        timestamped_entry = leaf_input.timestamped_entry
        print(timestamped_entry.entry_type)
        # print(timestamped_entry.extensions)

        print(hexdump(timestamped_entry.tdf))

        sct_version = rfc6962.Version(b'\x00')
        signature_type = rfc6962.SignatureType(b'\x00')

        # print(sct_version)
        # print(sct_version.tdf)
        # print(signature_type)
        # print(signature_type.tdf)

        signature_input = \
            sct_version.tdf + signature_type.tdf + timestamped_entry.tdf

        # print(hexdump(signature_input))

        pubkey = get_pubkey_pem(string_without_prefix('https://', ctlog_url))
        # print(pubkey)

        signature = decode_from_b64(signature_b64)

        verified, output, cmd_res = verify_signature(
            signature_input,
            signature,
            pubkey
        )
        print('--')
        print(verified)
        print('--')


def main():
    parser = create_parser()
    args = parser.parse_args()
    logger = setup_logging(args.loglevel)
    logger.debug(args)

    get_entry(
        index=args.entry, ctlog_url=args.log, signature_b64=args.signature)


if __name__ == '__main__':
    main()



import collections

from utlz import flo

from utlz import StructContext


_SctListEntry = collections.namedtuple(
    typename='SctListEntry',
    field_names=[
        'sct_len',
        'sct_der',
    ]
)


_TlsExtension18 = collections.namedtuple(
    typename='TlsExtension18',
    field_names=[
        'tls_extension_type',
        'tls_extension_len',
        'signed_certificate_timestamp_list_len',
        'sct_list',
    ]
)


def TlsExtension18(extension_18_tdf):
    with StructContext(extension_18_tdf) as struct:
        data_dict = {
            'tls_extension_type':                    struct.read('!H'),
            'tls_extension_len':                     struct.read('!H'),
            'signed_certificate_timestamp_list_len': struct.read('!H'),
        }
        sct_list = []
        while struct.offset < struct.length:
            sct_len = struct.read('!H')
            sct_der = struct.read(flo('!{sct_len}s'))
            sct_list.append(_SctListEntry(sct_len, sct_der))
        return _TlsExtension18(sct_list=sct_list, **data_dict)


_SignedCertificateTimestampList = collections.namedtuple(
    typename='SignedCertificateTimestampList',
    field_names=[
        'signed_certificate_timestamp_list_len',
        'sct_list',
    ]
)


def SignedCertificateTimestampList(sctlist):
    with StructContext(sctlist) as struct:
        data_dict = {
            'signed_certificate_timestamp_list_len': struct.read('!H'),
        }
        sct_list = []
        while struct.offset < struct.length:
            sct_len = struct.read('!H')
            sct_der = struct.read(flo('!{sct_len}s'))
            sct_list.append(_SctListEntry(sct_len, sct_der))
        return _SignedCertificateTimestampList(sct_list=sct_list, **data_dict)



'''Download, merge and summarize known logs for Certificate Transparency (CT).

Print output to stdout, warnings and errors to stderr.

Currently there exist three log lists with differing infos:

1. listing of webpage https://www.certificate-transparency.org/known-logs
2. log_list.json
3. all_logs_list.json.

This three log lists will be merged into one list in the future.  Discussion:
https://groups.google.com/forum/?fromgroups#!topic/certificate-transparency/zBv7EK0522w
'''

import argparse
import datetime
import json
import logging

from utlz import first_paragraph, flo, red

from ctutlz.ctlog import logs_dict_from_webpage, download_log_list
from ctutlz.ctlog import set_operator_names, print_schema, unset_operator_names
from ctutlz.ctlog import URL_LOG_LIST, URL_ALL_LOGS, Log, Logs, ChromeStates
from ctutlz.utils.logger import VERBOSE, init_logger, setup_logging, logger
from ctutlz._version import __version__


def create_parser():
    parser = argparse.ArgumentParser(description=first_paragraph(__doc__))
    parser.epilog = __doc__.split('\n', 1)[-1]
    parser.add_argument('-v', '--version',
                        action='version',
                        default=False,
                        version=__version__,
                        help='print version number')

    me1 = parser.add_mutually_exclusive_group()
    me1.add_argument('--short',
                     dest='loglevel',
                     action='store_const',
                     const=logging.INFO,
                     default=VERBOSE,  # default loglevel if nothing set
                     help='show short results')
    me1.add_argument('--debug',
                     dest='loglevel',
                     action='store_const',
                     const=logging.DEBUG,
                     help='show more for diagnostic purposes')

    me2 = parser.add_mutually_exclusive_group()
    me2.add_argument('--json',
                     action='store_true',
                     dest='print_json',
                     help='print merged log lists as json')
    me2.add_argument('--schema',
                     action='store_true',
                     dest='print_schema',
                     help='print json schema')
    return parser


def warn_inconsistency(url, val_a, val_b):

    # suppress warning doubles (i know it's hacky)
    key = flo('{url}' + ''.join(sorted(flo('{val_a}{val_b}'))))
    if not hasattr(warn_inconsistency, 'seen'):
        warn_inconsistency.seen = {}
    if not warn_inconsistency.seen.get(key, False):
        warn_inconsistency.seen[key] = True
    else:
        return

    logger.warn(red(flo(
        'inconsistent data for log {url}: {val_a} != {val_b}')))


def merge_logs(log_a, log_b):
    '''Return two logs merged into one log.

    If two values exist for one key and the differ, take value of log_a and
    print out a warning (to stderr).

    Args:
        log_a (ctutlz.ctlog.Log)
        log_b (ctutlz.ctlog.Log)

    Return: ctutlz.ctlog.Log
    '''
    logdict_a = log_a._asdict()
    logdict_b = log_b._asdict()

    merged = {}

    for key, val_a in logdict_a.items():
        val_b = logdict_b[key]
        if val_a is None:
            merged[key] = val_b  # may be None or not
        else:
            if val_b is not None and val_a != val_b:
                warn_inconsistency(log_a.url, val_a, val_b)
            merged[key] = val_a

    nc_a = log_a.id_b64_non_calculated
    nc_b = log_b.id_b64_non_calculated
    if nc_a is not None:
        if nc_b is not None and nc_a != nc_b:
            warn_inconsistency(log_a.url, nc_a, nc_b)
        if log_b.id_b64 is not None and nc_a != log_b.id_b64:
            warn_inconsistency(log_a.url, nc_a, log_b.id_b64)
    if nc_b is not None and log_a.id_b64 is not None:
        if log_a.id_b64 != nc_b:
            warn_inconsistency(log_a.url, log_a.id_b64, nc_b)

    return Log(**merged)


def merge_log_list_r(list_a, list_b, merged_logs=[], unmerged_a=[]):
    '''Merge two log lists `list_a` and `list_b` merged by log url.'''
    if len(list_a) == 0:
        unmerged_b = list_b
        return (merged_logs, unmerged_a, unmerged_b)

    log_a = list_a[0]
    rest_a = list_a[1:]

    matches = [item
               for item in list_b
               if item.url.rstrip('/') == log_a.url.rstrip('/')]
    rest_b = [item
              for item in list_b
              if item.url.rstrip('/') != log_a.url.rstrip('/')]

    merged_log = None
    unmerged_log = None

    if matches:
        merged_log = log_a
        for log_b in matches:
            merged_log = merge_logs(merged_log, log_b)
    else:
        unmerged_log = log_a

    if merged_log:
        merged_logs = merged_logs + [merged_log]
    if unmerged_log:
        unmerged_a = unmerged_a + [unmerged_log]

    return merge_log_list_r(rest_a, rest_b, merged_logs, unmerged_a)


def merge_enrich_a_with_b(log_list_a, log_list_b):
    '''
    Differing log values in log_list_a win over log_list_b.
    '''
    merged, rest_a, rest_b = merge_log_list_r(log_list_a, log_list_b)
    return merged + rest_a, rest_b


def merge_overwrite_a_with_b(log_list_a, log_list_b):
    '''
    Differing log values in log_list_b win over log_list_a.
    '''
    merged, rest_b, rest_a = merge_log_list_r(log_list_b, log_list_a)
    return merged + rest_a, rest_b


def merge_log_lists(included_from_webpage,
                    frozen_from_webpage,
                    pending_from_webpage,
                    disqualified_from_webpage,
                    rejected_from_webpage,
                    distrusted_from_webpage,
                    other_from_webpage,
                    special_from_webpage,

                    all_from_webpage,  # known-logs.html

                    log_list_logs,     # log_list.json
                    all_logs,          # all_logs_list.json
                    **_):
    '''Merge log lists, warn on log list errors and return merged logs.'''
    # log lists

    # log_list.json contains the logs which are chrome ct policy compliant
    #
    # 'll_...' means:     log is listed in log_list.json
    # 'nn_...' means: log is not listed in log_list.json

    ll_included = []
    ll_frozen = []
    ll_pending = []
    ll_disqualified = []
    ll_rejected = []
    ll_distrusted = []
    ll_other = []
    ll_special = []

    nn_included = []
    nn_frozen = []
    nn_pending = []
    nn_disqualified = []
    nn_rejected = []
    nn_distrusted = []
    nn_other = []
    nn_special = []

    # merge log_list.json with log lists from webpage

    ll_rest = log_list_logs

    ll_included, ll_rest, nn_included = \
        merge_log_list_r(ll_rest, included_from_webpage)
    ll_frozen, ll_rest, nn_frozen = \
        merge_log_list_r(ll_rest, frozen_from_webpage)
    ll_pending, ll_rest, nn_pending = \
        merge_log_list_r(ll_rest, pending_from_webpage)
    ll_disqualified, ll_rest, nn_rejected = \
        merge_log_list_r(ll_rest, disqualified_from_webpage)
    ll_rejected, ll_rest, nn_rejected = \
        merge_log_list_r(ll_rest, rejected_from_webpage)
    ll_distrusted, ll_rest, nn_distrusted = \
        merge_log_list_r(ll_rest, distrusted_from_webpage)
    ll_other, ll_rest, nn_other = \
        merge_log_list_r(ll_rest, other_from_webpage)
    ll_special, ll_rest, nn_special = \
        merge_log_list_r(ll_rest, special_from_webpage)

    # `ll_rest` now contains all logs from log_list.json which are not
    # listed on webpage (this list should be empty, else the webpage is missing
    # entries)

    # merge log lists with all_logs.json

    all_rest = all_logs  # logs listed in all_logs.json

    ll_included, all_rest = \
        merge_enrich_a_with_b(ll_included, all_rest)
    ll_frozen, all_rest = \
        merge_enrich_a_with_b(ll_frozen, all_rest)
    ll_pending, all_rest = \
        merge_enrich_a_with_b(ll_pending, all_rest)
    ll_disqualified, all_rest = \
        merge_enrich_a_with_b(ll_disqualified, all_rest)
    ll_rejected, all_rest = \
        merge_enrich_a_with_b(ll_rejected, all_rest)
    ll_distrusted, all_rest = \
        merge_enrich_a_with_b(ll_distrusted, all_rest)
    ll_other, all_rest = \
        merge_enrich_a_with_b(ll_other, all_rest)
    ll_special, all_rest = \
        merge_enrich_a_with_b(ll_special, all_rest)

    nn_included, all_rest = \
        merge_overwrite_a_with_b(nn_included, all_rest)
    nn_frozen, all_rest = \
        merge_overwrite_a_with_b(nn_frozen, all_rest)
    nn_pending, all_rest = \
        merge_overwrite_a_with_b(nn_pending, all_rest)
    nn_disqualified, all_rest = \
        merge_overwrite_a_with_b(nn_disqualified, all_rest)
    nn_rejected, all_rest = \
        merge_overwrite_a_with_b(nn_rejected, all_rest)
    nn_distrusted, all_rest = \
        merge_overwrite_a_with_b(nn_distrusted, all_rest)
    nn_other, all_rest = \
        merge_overwrite_a_with_b(nn_other, all_rest)
    nn_special, all_rest = \
        merge_overwrite_a_with_b(nn_special, all_rest)

    # currently only special purpose logs are listet on webpage known-logs.html
    #
    # # warn for missing logs on webpage
    #
    # for log in ll_rest:
    #     logger.warn(red(flo(
    #         'log in log_list.json not listet on webpage: {log.url}')))
    #
    # for log in all_rest:
    #     logger.warn(red(flo(
    #         'log in all_logs.json not listed on webpage: {log.url}')))

    # warn for wrongly listed logs

    for log in ll_pending:
        logger.warn(red(flo(
            'log pending for inclusion listed in log_list.json: {log.url}')))

    for log in ll_rejected:
        logger.warn(red(flo(
            'rejeted log listet in log_list.json: {log.url}')))

    for log in ll_distrusted:
        logger.warn(red(flo(
            'distrusted log listet in log_list.json: {log.url}')))

    for log in ll_other:
        logger.warn(red(flo(
            'other purpose log listet in log_list.json: {log.url}')))

    for log in nn_included:
        logger.warn(red(flo(
            'chrome included log not listet in log_list.json: {log.url}')))

    for log in nn_frozen:
        logger.warn(red(flo(
            'chrome frozen log not listet in log_list.json: {log.url}')))

    ll_rest, all_rest = \
        merge_enrich_a_with_b(ll_rest, all_rest)
    rest = ll_rest + all_rest

    # warn for logs only listed on webpage

    _, webpage_rest, _ = merge_log_list_r(all_from_webpage, log_list_logs)
    _, webpage_rest, _ = merge_log_list_r(all_from_webpage, all_logs)
    for log in webpage_rest:
        logger.warn(red(flo(
            'log not listet in log_list.json nor all_logs.json: {log.url}')))

    return [
        {'heading': 'included logs (log_list.json, webpage)',
         'logs': ll_included},
        {'heading': 'frozen logs (log_list.json, webpage)',
         'logs': ll_frozen},
        {'heading': 'pending logs (log_list.json, webpage)',
         'logs': ll_pending},
        {'heading': 'disqualified logs (log_list.json, webpage)',
         'logs': ll_disqualified},
        {'heading': 'rejected logs (log_list.json, webpage)',
         'logs': ll_rejected},
        {'heading': 'distrusted logs (log_list.json, webpage)',
         'logs': ll_distrusted},

        {'heading': 'included logs NOT IN log_list.json '
                    '(webpage, all_logs.json)',
         'logs': nn_included},
        {'heading': 'frozen logs NOT IN log_list.json (webpage, all_logs.json)',
         'logs': nn_frozen},
        {'heading': 'pending logs (webpage, all_logs.json)',
         'logs': nn_pending},
        {'heading': nn_disqualified,
         'logs': nn_disqualified},
        {'heading': 'rejected logs (webpage, all_logs.json)',
         'logs': nn_rejected},
        {'heading': 'distrusted logs (webpage, all_logs.json)',
         'logs': nn_distrusted},
        {'heading': 'other logs (webpage, all_logs.json)',
         'logs': nn_other},
        {'heading': 'special purpose logs (webpage, all_logs.json)',
         'logs': nn_special},

        {'heading': 'UNLISTED ON WEBPAGE (log_list.json or all_logs.json)',
         'logs': rest},
    ]


def data_structure_from_log(log):
    log_data = dict(log._asdict())

    log_data['id_b64'] = log.id_b64
    log_data['pubkey'] = log.pubkey
    log_data['scts_accepted_by_chrome'] = \
        log.scts_accepted_by_chrome

    return log_data


def list_from_lists(log_lists):
    log_list = []
    for item_dict in log_lists:
        for log in item_dict['logs']:
            log_data = data_structure_from_log(log)
            log_list.append(log_data)
    return log_list


def show_log(log, order=3):
    logger.verbose('#'*order + flo(' {log.url}\n'))

    logdict = log._asdict()

    for key, value in logdict.items():
        if key == 'id_b64_non_calculated' and value == log.id_b64:
            value = None  # don't log this value
        if key == 'operated_by':
            value = ', '.join(value)
        # avoid markdown syntax interpretation and improve readablity
        key = key.replace('_', ' ')
        if value is not None:
            logger.verbose(flo('* __{key}__: `{value}`'))

    logger.verbose(flo('* __scts accepted by chrome__: '
                       '{log.scts_accepted_by_chrome}'))

    if log.key is not None:
        logger.verbose(flo('* __id b64__: `{log.id_b64}`'))
        logger.verbose(flo('* __pubkey__:\n```\n{log.pubkey}\n```'))

    logger.verbose('')


def show_logs(logs, heading, order=2):
    if len(logs) > 0:
        logger.info('#'*order + flo(' {heading}\n'))
        s_or_not = 's'
        if len(logs) == 1:
            s_or_not = ''
        # show log size
        logger.info('%i log%s\n' % (len(logs), s_or_not))

        # list log urls
        for log in logs:
            if logger.level < logging.INFO:
                anchor = log.url.replace('/', '')
                logger.verbose(flo('* [{log.url}](#{anchor})'))
            else:
                logger.info(flo('* {log.url}'))
        logger.info('')
        for log in logs:
            show_log(log)


def ctloglist(print_json=None):
    '''Gather ct-log lists and print the merged log list.

    Args:
        print_json(boolean): If True, print merged log list as json data.
                             Else print as markdown.
    '''
    if not print_json:
        today = datetime.date.today()
        now = datetime.datetime.now()

        logger.info('# Known Certificate Transparency (CT) Logs\n')
        logger.verbose('Created with [ctloglist]'
                       '(https://github.com/theno/ctutlz#ctloglist)\n')
        logger.verbose('Merged log lists:')
        logger.verbose("* webpage [known logs]"
                       '(https://www.certificate-transparency.org/known-logs)')
        logger.verbose('* [log_list.json]'
                       '(https://www.gstatic.com/ct/log_list/log_list.json)')
        logger.verbose('* [all_logs_list.json]('
                       'https://www.gstatic.com/ct/log_list/all_logs_list.json)'
                       '\n')
        logger.info(flo('Version (Date): {today}\n'))
        logger.verbose(flo('Datetime: {now}\n'))
        logger.info('')  # formatting: insert empty line

    # from webpage

    webpage_dict = logs_dict_from_webpage()

    all_from_webpage = Logs([log_dict
                             for log_list
                             in [webpage_dict[key] for key in webpage_dict]
                             for log_dict
                             in log_list])

    included_from_webpage = Logs(webpage_dict.get('included_in_chrome', []))
    try:
        webpage_dict.pop('included_in_chrome')
    except KeyError:
        pass

    frozen_from_webpage = Logs(webpage_dict.get('frozen_logs', []))
    try:
        webpage_dict.pop('frozen_logs')
    except KeyError:
        pass

    pending_from_webpage = Logs(webpage_dict.get('pending_inclusion_in_chrome',
                                                 []))
    try:
        webpage_dict.pop('pending_inclusion_in_chrome')
    except KeyError:
        pass

    disqualified_from_webpage = \
        Logs(webpage_dict.get('disqualified_from_chrome', []))
    try:
        webpage_dict.pop('disqualified_from_chrome')
    except KeyError:
        pass

    rejected_from_webpage = Logs(webpage_dict.get('rejected_by_chrome', []))
    try:
        webpage_dict.pop('rejected_by_chrome')
    except KeyError:
        pass

    distrusted_from_webpage = Logs(webpage_dict.get(
        'completely_distrusted_by_chrome', []))
    try:
        webpage_dict.pop('completely_distrusted_by_chrome')
    except KeyError:
        pass

    other_from_webpage = Logs(webpage_dict.get('other_logs', []))
    try:
        webpage_dict.pop('other_logs')
    except KeyError:
        pass

    special_from_webpage = Logs(webpage_dict.get('special_purpose_logs', []))
    try:
        webpage_dict.pop('special_purpose_logs')
    except KeyError:
        pass

    unknown_log_titles = [key for key in webpage_dict.keys()]
    if unknown_log_titles:
        logger.error(red(flo(
            'unknown log titles (i.e. log states): {unknown_log_titles}')))

    # log_list.json: chrome ct policy compliant logs

    log_list_dict = download_log_list(URL_LOG_LIST)
    set_operator_names(log_list_dict)
    for log_dict in log_list_dict['logs']:
        if 'disqualified_at' in log_dict.keys():
            log_dict['chrome_state'] = ChromeStates.DISQUALIFIED
        elif 'final_sth' in log_dict.keys():
            log_dict['chrome_state'] = ChromeStates.FROZEN
        else:
            log_dict['chrome_state'] = ChromeStates.INCLUDED

    log_list_logs = Logs(log_list_dict['logs'])

    # all_logs_list.json

    all_dict = download_log_list(URL_ALL_LOGS)
    set_operator_names(all_dict)

    all_logs = Logs(all_dict['logs'])

    # merge lists and show the result

    log_lists = merge_log_lists(**locals())

    if print_json:

        data = {
            'operators': all_dict['operators'],
            'logs': list_from_lists(log_lists)
        }
        unset_operator_names(data)
        json_str = json.dumps(data, indent=4, sort_keys=True)
        print(json_str)

    else:
        for item in log_lists:
            show_logs(item['logs'], item['heading'])


def main():
    init_logger()
    parser = create_parser()
    args = parser.parse_args()
    setup_logging(args.loglevel)
    logger.debug(args)
    if args.print_schema:
        print_schema()
    else:
        ctloglist(args.print_json)


if __name__ == '__main__':
    main()



import collections

from cryptography.hazmat.backends.openssl.backend import backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec, dsa, rsa
from OpenSSL.crypto import verify, X509, PKey, Error as OpenSSL_crypto_Error


SctVerificationResult = collections.namedtuple(
    typename='SctVerificationResult',
    field_names=[
        'ee_cert',   # DER format
        'sct',       # type: Sct
        'log',       # type: Log
        'verified',  # True or False
    ]
)


def find_log(sct, logs):
    for log in logs:
        if log.id_der == sct.log_id.tdf:
            return log
    return None


def pkey_from_cryptography_key(crypto_key):
    '''
    Modified version of `OpenSSL.crypto.PKey.from_cryptography_key()` of
    PyOpenSSL which also accepts EC Keys
    (cf. https://github.com/pyca/pyopenssl/pull/636).
    '''
    pkey = PKey()
    if not isinstance(crypto_key, (rsa.RSAPublicKey, rsa.RSAPrivateKey,
                                   dsa.DSAPublicKey, dsa.DSAPrivateKey,
                                   ec.EllipticCurvePublicKey,
                                   ec.EllipticCurvePrivateKey)):
        raise TypeError("Unsupported key type")

    pkey._pkey = crypto_key._evp_pkey
    if isinstance(crypto_key, (rsa.RSAPublicKey, dsa.DSAPublicKey,
                               ec.EllipticCurvePublicKey)):
        pkey._only_public = True
    pkey._initialized = True
    return pkey


def verify_signature(signature_input, signature,
                     pubkey_pem, digest_algo='sha256'):
    '''Verify if `signature` over `signature_input` was created using
    `digest_algo` by the private key of the `pubkey_pem`.

    A signature is the private key encrypted hash over the signature input data.

    Args:
        signature_input(bytes): signed data
        signature(bytes):
        pubkey_pem(str): PEM formatted pubkey
        digest_algo(str): name of the used digest hash algorithm
                          (default: 'sha256')

    Return:
        True, if signature could be verified
        False, else
    '''
    cryptography_key = serialization.load_pem_public_key(pubkey_pem, backend)
    pkey = pkey_from_cryptography_key(cryptography_key)

    auxiliary_cert = X509()
    auxiliary_cert.set_pubkey(pkey)

    try:
        verify(cert=auxiliary_cert, signature=signature,
               data=signature_input, digest=digest_algo)
    except OpenSSL_crypto_Error:
        return False

    return True


def verify_sct(ee_cert, sct, logs,
               issuer_cert, more_issuer_cert_candidates,
               sign_input_func):
    log = find_log(sct, logs)
    if log:
        verified = verify_signature(
            signature_input=sign_input_func(ee_cert, sct, issuer_cert),
            signature=sct.signature,
            pubkey_pem=log.pubkey.encode('ascii')
        )

        if not verified and more_issuer_cert_candidates is not None:
            # Sometimes the certificate chain is disordered (this is only
            # relevant in case of of sct-by-cert, because the signature input
            # then also depends by the issuer key hash)

            # Try to verify against all other certs of the chain (candidates).
            for issuer_cert in more_issuer_cert_candidates:
                verified = verify_signature(
                    signature_input=sign_input_func(ee_cert, sct, issuer_cert),
                    signature=sct.signature,
                    pubkey_pem=log.pubkey.encode('ascii')
                )
                if verified:
                    break

        # # TODO DEVEL
        # print('CREATE TEST DATA FOR test_verify_sct()')
        # open('signature_input.bin', 'wb').write(sign_input_func(ee_cert,
        #                                                         sct,
        #                                                         issuer_cert))
        # open('signature.der', 'wb').write(sct.signature)
        # open('pubkey.pem', 'wb').write(log.pubkey.encode('ascii'))

        return SctVerificationResult(ee_cert, sct, log, verified)
    return SctVerificationResult(ee_cert, sct, log=None, verified=False)


def verify_scts(ee_cert, scts, logs,
                issuer_cert, more_issuer_cert_candidates,
                sign_input_func):
    if scts:
        return [verify_sct(ee_cert, sct, logs,
                           issuer_cert, more_issuer_cert_candidates,
                           sign_input_func)
                for sct
                in scts]
    return []



import base64
import hashlib


def encode_to_b64(arg):
    '''Return arg as Base64 encoded string (not as bytes-str).'''
    res = base64.b64encode(arg)
    # assert type(res) == bytes, 'type(result) is ' + str(type(res))
    return res.decode('ascii')


def decode_from_b64(arg):
    return base64.b64decode(arg)


def sha256_digest(arg):
    return hashlib.sha256(arg).digest()


def digest_from_b64(arg):
    return sha256_digest(decode_from_b64(arg))


def digest_from_b64_encoded_to_b64(arg):
    return encode_to_b64(digest_from_b64(arg))



'''Compile cffi loader in order to use OpenSSL c-code functionality to register
a callback for TLS extension 18 results in the SSL context object created
with PyOpenSSL.

CFFI will be used in API level:
https://cffi.readthedocs.io/en/latest/overview.html#real-example-api-level-out-of-line

The callback is written in Python:
https://cffi.readthedocs.io/en/latest/using.html#extern-python-new-style-callbacks

FFI means foreign function interface:
https://enwikipedia.org/wiki/Foreign_function_interface
'''

from cffi import FFI


def create_ffibuilder():
    module_name = 'ctutlz.tls.handshake_openssl'

    libraries = ['ssl', 'crypto']

    csource_ffi = r'''
        #include "stdio.h"  // FILE

        #include "openssl/bio.h"
        #include "openssl/err.h"
        #include "openssl/ssl.h"
    '''

    cdefinitions_lib = r'''
        // for TLS extension 18

        typedef struct ssl_ctx_st SSL_CTX;
        typedef struct ssl_st SSL;

        typedef int (*custom_ext_add_cb) (SSL *s, unsigned int ext_type,
                                          const unsigned char **out,
                                          size_t *outlen, int *al,
                                          void *add_arg);
        typedef void (*custom_ext_free_cb) (SSL *s, unsigned int ext_type,
                                            const unsigned char *out,
                                            void *add_arg);
        typedef int (*custom_ext_parse_cb) (SSL *s, unsigned int ext_type,
                                            const unsigned char *in,
                                            size_t inlen, int *al,
                                            void *parse_arg);
        int SSL_CTX_add_client_custom_ext(SSL_CTX *ctx, unsigned int ext_type,
                                          custom_ext_add_cb add_cb,
                                          custom_ext_free_cb free_cb,
                                          void *add_arg,
                                          custom_ext_parse_cb parse_cb,
                                          void *parse_arg);

        extern "Python" static int serverinfo_cli_parse_cb(SSL *s,
                                                           unsigned int ext_type,
                                                           const unsigned char *in,
                                                           size_t inlen,
                                                           int *al, void *arg);
    '''

    ffibuilder = FFI()
    ffibuilder.set_source(module_name, csource_ffi, libraries=libraries)
    ffibuilder.cdef(cdefinitions_lib)
    return ffibuilder


# hook for the kwarg 'cffi_modules' of the setup() call in setup.py
ffibuilder = create_ffibuilder()


# hook for '__main__' clause in verify_scts.py
def compile():
    ffibuilder.compile(verbose=False)



"""Certificate Transparency utils library and scripts.

* https://github.com/theno/ctutlz
* https://pypi.python.org/pypi/ctutlz
"""

import os
import shutil
from setuptools import setup, find_packages
from codecs import open


def create_readme_with_long_description():
    this_dir = os.path.abspath(os.path.dirname(__file__))
    readme_md = os.path.join(this_dir, 'README.md')
    readme = os.path.join(this_dir, 'README')
    if os.path.isfile(readme_md):
        if os.path.islink(readme):
            os.remove(readme)
        shutil.copy(readme_md, readme)
    try:
        import pypandoc
        long_description = pypandoc.convert(readme_md, 'rst', format='md')
        if os.path.islink(readme):
            os.remove(readme)
        with open(readme, 'w') as out:
            out.write(long_description)
    except(IOError, ImportError, RuntimeError):
        if os.path.isfile(readme_md):
            os.remove(readme)
            os.symlink(readme_md, readme)
        with open(readme, encoding='utf-8') as in_:
            long_description = in_.read()
    return long_description


this_dir = os.path.abspath(os.path.dirname(__file__))
filename = os.path.join(this_dir, 'ctutlz', '_version.py')
with open(filename, 'rt') as fh:
    version = fh.read().split('"')[1]

description = __doc__.split('\n')[0]
long_description = create_readme_with_long_description()

setup(
    name='ctutlz',
    version=version,
    description=description,
    long_description=long_description,
    url='https://github.com/theno/ctutlz',
    author='Theodor Nolte',
    author_email='ctutlz@theno.eu',
    license='MIT',
    entry_points={
        'console_scripts': [
            'ctloglist = ctutlz.scripts.ctloglist:main',
            'decompose-cert = ctutlz.scripts.decompose_cert:main',
            'verify-scts = ctutlz.scripts.verify_scts:main',
        ],
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='python development utilities library '
             'certificate-transparency ct signed-certificate-timestamp sct',
    packages=find_packages(exclude=[
        'contrib',
        'docs',
        'tests',
    ]),
    package_data={'ctutlz': ['really_all_logs.json', 'log_list_schema.json'], },
    install_requires=[
        'cffi>=1.4.0',
        'cryptography<2.0.0',
        'html2text>=2016.9.19',
        'pyasn1>=0.2.0,<0.3.0',
        'pyasn1-modules>=0.0.9,<0.2.0',
        'pyOpenSSL>=17.0.0,<=17.3.0',
        'requests>=2.17.0',
        'utlz>=0.9.0',
    ],
    extras_require={
        'dev': ['pypandoc'],
    },
    setup_requires=[
        'cffi>=1.4.0'
    ],
    cffi_modules=[
        'ctutlz/tls/handshake_openssl_build.py:ffibuilder'
    ],
)
